from re import A
import unittest
import csv
import ast
import sys
import json
import os
import numpy as np

sys.path.append("./src")
from chemsynthcalc import ChemicalReaction
from chemsynthcalc.chem_errors import (
    BadCoeffiecients,
    NoSuchMode,
    NoSeparator,
    InvalidCharacter,
    ReactantProductDifference,
    ReactionNotBalanced,
    NoSuchAlgorithm,
)


class TestChemicalReaction(unittest.TestCase):
    def setUp(self) -> None:
        dir_name = os.path.dirname(os.path.realpath(__file__))
        with open(dir_name+"/testing_reactions.csv") as csvfile:
            reader = list(csv.reader(csvfile))[1:]
        self.reactions_set = [(i[0], ast.literal_eval(i[1])) for i in reader]

        self.test_data = [
            {
                "initial reaction": "H2+O2==H2O",
                "reaction matrix": "[[2. 0. 2.]\n [0. 2. 1.]]",
                "mode": "balance",
                "formulas": ["H2", "O2", "H2O"],
                "coefficients": [2, 1, 2],
                "normalized coefficients": [1, 0.5, 1],
                "algorithm": "inverse",
                "is balanced": True,
                "final reaction": "2H2+O2==2H2O",
                "final reaction normalized": "H2+0.5O2==H2O",
                "molar masses": [2.016, 31.998, 18.015],
                "target": "H2O",
                "masses": [0.11190674, 0.88809326, 1.0],
            },
            {
                "initial reaction": "P+HNO3+H2O<->H3PO4+NO",
                "reaction matrix": "[[1. 0. 0. 1. 0.]\n [0. 1. 2. 3. 0.]\n [0. 1. 0. 0. 1.]\n [0. 3. 1. 4. 1.]]",
                "mode": "balance",
                "formulas": ["P", "HNO3", "H2O", "H3PO4", "NO"],
                "coefficients": [3, 5, 2, 3, 5],
                "normalized coefficients": [1, 1.66666667, 0.66666667, 1, 1.66666667],
                "algorithm": "inverse",
                "is balanced": True,
                "final reaction": "3P+5HNO3+2H2O<->3H3PO4+5NO",
                "final reaction normalized": "P+1.66666667HNO3+0.66666667H2O<->H3PO4+1.66666667NO",
                "molar masses": [30.974, 63.012, 18.015, 97.994, 30.006],
                "target": "H3PO4",
                "masses": [0.31608058, 1.07169827, 0.12255852, 1.0, 0.51033737],
            },
            {
                "initial reaction": "Bi2O3+TiO2→Bi4Ti3O12",
                "reaction matrix": "[[ 2.  0.  4.]\n [ 3.  2. 12.]\n [ 0.  1.  3.]]",
                "mode": "balance",
                "formulas": ["Bi2O3", "TiO2", "Bi4Ti3O12"],
                "coefficients": [2, 3, 1],
                "normalized coefficients": [2, 3, 1],
                "algorithm": "inverse",
                "is balanced": True,
                "final reaction": "2Bi2O3+3TiO2→Bi4Ti3O12",
                "final reaction normalized": "2Bi2O3+3TiO2→Bi4Ti3O12",
                "molar masses": [465.957, 79.865, 1171.509],
                "target": "Bi4Ti3O12",
                "masses": [0.79548172, 0.20451828, 1.0],
            },
            {
                "initial reaction": "CaSO4*2H2O⇄CaSO4*0.5H2O+H2O",
                "reaction matrix": "[[1.  1.  0. ]\n [1.  1.  0. ]\n [6.  4.5 1. ]\n [4.  1.  2. ]]",
                "mode": "balance",
                "formulas": ["CaSO4*2H2O", "CaSO4*0.5H2O", "H2O"],
                "coefficients": [2, 2, 3],
                "normalized coefficients": [1, 1, 1.5],
                "algorithm": "general pseudoinverse",
                "is balanced": True,
                "final reaction": "2CaSO4*2H2O⇄2CaSO4*0.5H2O+3H2O",
                "final reaction normalized": "CaSO4*2H2O⇄CaSO4*0.5H2O+1.5H2O",
                "molar masses": [172.164, 145.1415, 18.015],
                "target": "CaSO4*0.5H2O",
                "masses": [1.18618038, 1.0, 0.18618038],
            },
        ]

    # set of tests for reaction string validity checks
    def test_empty_reaction(self) -> None:
        string = ""
        self.assertRaises(ValueError, lambda: ChemicalReaction(string))

    def test_invalid_character(self) -> None:
        string = "H2+O2=фH2O"
        self.assertRaises(InvalidCharacter, lambda: ChemicalReaction(string))

    def test_invalid_string(self) -> None:
        string = "H2O2H2O"
        self.assertRaises(NoSeparator, lambda: ChemicalReaction(string))

    def test_wrong_reactant_separator(self) -> None:
        string = "H2&O2=H2O"
        self.assertRaises(InvalidCharacter, lambda: ChemicalReaction(string))

    def test_wrong_separator(self) -> None:
        string = "H2+O2←H2O"
        self.assertRaises(InvalidCharacter, lambda: ChemicalReaction(string))

    def test_unbalanceble_reaction(self) -> None:
        string = "Ho+O2=H2O"
        self.assertRaises(
            ReactantProductDifference, lambda: ChemicalReaction(string).coefficients
        )

    # set of tests for ChemicalReaction with wrong parameters
    def test_reaction_type(self) -> None:
        string = 3
        self.assertRaises(TypeError, lambda: ChemicalReaction(string))

    def test_target_type(self) -> None:
        string = "Al2S3+HNO3<>S+NO2+Al(NO3)3+H2O"
        self.assertRaises(TypeError, lambda: ChemicalReaction(string, target="S"))

    def test_mode_type(self) -> None:
        string = "Al2S3+HNO3<>S+NO2+Al(NO3)3+H2O"
        self.assertRaises(TypeError, lambda: ChemicalReaction(string, mode=3))

    def test_target_mass_type(self) -> None:
        string = "Al2S3+HNO3<>S+NO2+Al(NO3)3+H2O"
        self.assertRaises(
            TypeError, lambda: ChemicalReaction(string, target_mass="3.13")
        )

    def test_rounding_order_type(self) -> None:
        string = "Al2S3+HNO3->S+NO2+Al(NO3)3+H2O"
        self.assertRaises(
            TypeError, lambda: ChemicalReaction(string, rounding_order="8")
        )

    def test_try_comb_type(self) -> None:
        string = "Al2S3+HNO3>S+NO2+Al(NO3)3+H2O"
        self.assertRaises(TypeError, lambda: ChemicalReaction(string, try_comb=1))

    def test_balance_algorithm_type(self) -> None:
        string = "Al2S3+HNO3>S+NO2+Al(NO3)3+H2O"
        self.assertRaises(
            TypeError, lambda: ChemicalReaction(string).balance_reaction(algorithm=3)
        )

    def test_balance_intify_type(self) -> None:
        string = "Al2S3+HNO3>S+NO2+Al(NO3)3+H2O"
        self.assertRaises(
            TypeError, lambda: ChemicalReaction(string).balance_reaction(intify=3)
        )

    def test_balance_max_comb_type(self) -> None:
        string = "Al2S3+HNO3>S+NO2+Al(NO3)3+H2O"
        self.assertRaises(
            TypeError, lambda: ChemicalReaction(string).balance_reaction(max_comb="10")
        )

    def test_balance_max_comb_less_than_zero(self) -> None:
        string = "Al2S3+HNO3>S+NO2+Al(NO3)3+H2O"
        self.assertRaises(
            ValueError, lambda: ChemicalReaction(string).balance_reaction(max_comb=-2)
        )

    def test_target_less_than_zero(self) -> None:
        string = "Al2S3+HNO3>S+NO2+Al(NO3)3+H2O"
        self.assertRaises(ValueError, lambda: ChemicalReaction(string, target=-1))

    def test_target_more_than_number_of_compounds(self) -> None:
        string = "Al2S3+HNO3->S+NO2+Al(NO3)3+H2O"
        self.assertRaises(ValueError, lambda: ChemicalReaction(string, target=4))

    def test_wrong_mode(self) -> None:
        string = "Al2S3+HNO3=S+NO2+Al(NO3)3+H2O"
        self.assertRaises(NoSuchMode, lambda: ChemicalReaction(string, mode="calc"))

    def test_target_mass_zero_or_less(self) -> None:
        string = "Al2S3+HNO3=S+NO2+Al(NO3)3+H2O"
        self.assertRaises(ValueError, lambda: ChemicalReaction(string, target_mass=0))

    def test_target_mass_zero_or_less(self) -> None:
        string = "Al2S3+HNO3=S+NO2+Al(NO3)3+H2O"
        self.assertRaises(ValueError, lambda: ChemicalReaction(string, target_mass=0))

    def test_rounding_order_zero_or_less(self) -> None:
        string = "Al2S3+HNO3=S+NO2+Al(NO3)3+H2O"
        self.assertRaises(
            ValueError, lambda: ChemicalReaction(string, rounding_order=0)
        )

    # set of tests for ChemicalReaction properties
    def test_reaction_matrix(self) -> None:
        for item in self.test_data:
            self.reaction = ChemicalReaction(item.get("initial reaction"))
            self.assertEqual(
                np.array2string(self.reaction.matrix), item.get("reaction matrix")
            )

    def test_mode(self) -> None:
        for item in self.test_data:
            self.reaction = ChemicalReaction(item.get("initial reaction"))
            self.assertEqual(self.reaction.mode, item.get("mode"))

    def test_formulas(self) -> None:
        for item in self.test_data:
            self.reaction = ChemicalReaction(item.get("initial reaction"))
            self.assertEqual(
                [str(formula) for formula in self.reaction.formulas],
                item.get("formulas"),
            )

    def test_coefficients(self) -> None:
        for item in self.test_data:
            self.reaction = ChemicalReaction(item.get("initial reaction"))
            self.assertEqual(self.reaction.coefficients, item.get("coefficients"))

    def test_normalized_coefficients(self) -> None:
        for item in self.test_data:
            self.reaction = ChemicalReaction(item.get("initial reaction"))
            self.assertEqual(
                self.reaction.normalized_coefficients,
                item.get("normalized coefficients"),
            )

    def test_algorithm(self) -> None:
        for item in self.test_data:
            self.reaction = ChemicalReaction(item.get("initial reaction"))
            self.reaction.coefficients
            self.assertEqual(self.reaction.algorithm, item.get("algorithm"))

    def test_is_balanced(self) -> None:
        for item in self.test_data:
            self.reaction = ChemicalReaction(item.get("initial reaction"))
            self.assertEqual(self.reaction.is_balanced, item.get("is balanced"))

    def test_final_reaction(self) -> None:
        for item in self.test_data:
            self.reaction = ChemicalReaction(item.get("initial reaction"))
            self.assertEqual(self.reaction.final_reaction, item.get("final reaction"))

    def test_final_reaction_normalized(self) -> None:
        for item in self.test_data:
            self.reaction = ChemicalReaction(item.get("initial reaction"))
            self.assertEqual(
                self.reaction.final_reaction_normalized,
                item.get("final reaction normalized"),
            )

    def test_molar_masses(self) -> None:
        for item in self.test_data:
            self.reaction = ChemicalReaction(item.get("initial reaction"))
            self.assertEqual(self.reaction.molar_masses, item.get("molar masses"))

    def test_target(self) -> None:
        for item in self.test_data:
            self.reaction = ChemicalReaction(item.get("initial reaction"))
            self.assertEqual(
                str(self.reaction.formulas[self.reaction.target]), item.get("target")
            )

    def test_masses(self) -> None:
        for item in self.test_data:
            self.reaction = ChemicalReaction(item.get("initial reaction"))
            self.assertEqual(self.reaction.masses, item.get("masses"))

    # test for different calculation modes
    def test_force_mode(self) -> None:
        string = "H2+O2=2H2O"
        self.reaction = ChemicalReaction(string, mode="force")
        self.assertEqual(self.reaction.masses, [0.05595337, 0.88809326, 1.0])

    def test_check_mode_fail(self) -> None:
        string = "H2+O2=2H2O"
        self.reaction = ChemicalReaction(string, mode="check")
        self.assertRaises(ReactionNotBalanced, lambda: self.reaction.coefficients)

    def test_check_mode(self) -> None:
        string = "2H2+O2=2H2O"
        self.reaction = ChemicalReaction(string, mode="check")
        self.assertEqual(self.reaction.masses, [0.11190674, 0.88809326, 1.0])

    def test_balance_method(self) -> None:
        for reaction in self.reactions_set[:107]:
            chemical_reaction = ChemicalReaction(reaction[0], mode="balance")
            self.assertEqual(chemical_reaction.coefficients, reaction[1])

    def test_feeding_coefs_with_zero(self) -> None:
        string = "H2+O2=H2O"
        self.reaction = ChemicalReaction(string, mode="balance")
        self.reaction.coefficients = [0, 1, 2]
        self.assertRaises(BadCoeffiecients, lambda: self.reaction.masses)

    def test_feeding_negative_coefs(self) -> None:
        string = "H2+O2=H2O"
        self.reaction = ChemicalReaction(string, mode="balance")
        self.reaction.coefficients = [-1, 1, 2]
        self.assertRaises(BadCoeffiecients, lambda: self.reaction.masses)

    def test_feeding_wrong_shape_coefs(self) -> None:
        string = "H2+O2=H2O"
        self.reaction = ChemicalReaction(string, mode="balance")
        self.reaction.coefficients = [1, 2]
        self.assertRaises(BadCoeffiecients, lambda: self.reaction.masses)

    # set of tests for reaction balancing methods
    def test_wrong_algorithm(self) -> None:
        chemical_reaction = ChemicalReaction("H2+O2=H2O")
        self.assertRaises(
            NoSuchAlgorithm, lambda: chemical_reaction.balance_reaction(algorithm="ginv")
            )

    def test_inverse_method(self) -> None:
        for reaction in self.reactions_set[:100]:
            chemical_reaction = ChemicalReaction(reaction[0])
            self.assertEqual(
                chemical_reaction.balance_reaction(algorithm="inv"), reaction[1]
            )

    def test_general_pseudoinverse_method(self) -> None:
        for reaction in self.reactions_set[101:107]:
            chemical_reaction = ChemicalReaction(reaction[0])
            self.assertEqual(
                chemical_reaction.balance_reaction(algorithm="gpinv"), reaction[1]
            )

    def test_partial_pseudoinverse_method(self) -> None:
        for reaction in self.reactions_set[108:110]:
            chemical_reaction = ChemicalReaction(reaction[0])
            coefs = chemical_reaction.balance_reaction(algorithm="ppinv", intify=False)
            coefs = [round(x, 10) for x in coefs]
            self.assertEqual(coefs, reaction[1])

    def test_comb_method(self) -> None:
        for reaction in self.reactions_set[111:113]:
            chemical_reaction = ChemicalReaction(reaction[0])
            coefs = chemical_reaction.balance_reaction(algorithm="comb")
            self.assertEqual(coefs, reaction[1])

    # exports tests
    def cleanup_files(self) -> None:
        listdir = os.listdir()
        for item in listdir:
            if item == "CSC_reaction_test.txt" or item == "CSC_reaction_test.json":
                os.remove(item)
        return

    def test_txt_export(self) -> None:
        string = self.test_data[0].get("initial reaction")
        self.formula = ChemicalReaction(string)
        self.formula.export_to_txt(filename="CSC_reaction_test.txt")

    def test_json_serialization(self) -> None:
        string = self.test_data[0].get("initial reaction")
        self.formula = ChemicalReaction(string)
        dictionary = json.loads(self.formula.as_json())
        output = {
            "initial reaction": "H2+O2==H2O",
            "reaction matrix": "[[2. 0. 2.]\n [0. 2. 1.]]",
            "mode": "balance",
            "formulas": ["H2", "O2", "H2O"],
            "coefficients": [2, 1, 2],
            "normalized coefficients": [1, 0.5, 1],
            "algorithm": "inverse",
            "is balanced": True,
            "final reaction": "2H2+O2==2H2O",
            "final reaction normalized": "H2+0.5O2==H2O",
            "molar masses": [2.016, 31.998, 18.015],
            "target": "H2O",
            "masses": [0.1119, 0.8881, 1.0],
        }
        self.assertEqual(dictionary, output)

    def test_json_export(self) -> None:
        string = self.test_data[0].get("initial reaction")
        self.formula = ChemicalReaction(string)
        self.formula.export_to_json(filename="CSC_reaction_test.json")

    def tearDown(self) -> None:
        self.cleanup_files()
        return super().tearDown()


if __name__ == "__main__":
    unittest.main()



"""
Module that contains custom errors for use in `ChemSynthCalc`
"""


class NoSuchAtom(Exception):
    """
    Found atom(s) that are not in the periodic table.
    """

    pass


class InvalidCharacter(Exception):
    """
    Found some characters that does not belong in
    chemical formula or reaction.
    """

    pass


class MoreThanOneAdduct(Exception):
    """
    There are more than one adduct (*).
    """

    pass


class BracketsNotPaired(Exception):
    """
    Some brackets are not come in pairs.
    """

    pass


class NoSuchMode(Exception):
    """
    Invalid calculation mode detected.
    """

    pass


class NoSuchAlgorithm(Exception):
    """
    Invalid calculation algorithm detected.
    """

    pass


class NoSeparator(Exception):
    """
    No separator found in the reaction string.
    """

    pass


class ReactionNotBalanced(Exception):
    """
    This reaction is not balanced.
    """

    pass


class ReactantProductDifference(Exception):
    """
    The elements in reaction are not evenly
    distributed in reactants and products:
    some of atoms are only in one part of
    reaction.
    """

    pass


class BadCoeffiecients(Exception):
    """
    The coefficients are not not compliant
    (they have no physical meaning).
    """

    pass



import imp
import unittest
import os
import sys
import json

sys.path.append("./src")
from chemsynthcalc import ChemicalFormula
from chemsynthcalc.chem_errors import (
    NoSuchAtom,
    InvalidCharacter,
    MoreThanOneAdduct,
    BracketsNotPaired,
)


class TestChemicalFormula(unittest.TestCase):
    def setUp(self) -> None:
        self.test_data = [
            {
                "formula": "H2O",
                "parsed formula": {"H": 2.0, "O": 1.0},
                "molar mass": 18.015,
                "mass percent": {"H": 11.19067444, "O": 88.80932556},
                "atomic percent": {"H": 66.66666667, "O": 33.33333333},
                "oxide percent": {"H2O": 100.0},
            },
            {
                "formula": "K2SO4",
                "parsed formula": {"K": 2.0, "S": 1.0, "O": 4.0},
                "molar mass": 174.252,
                "mass percent": {"K": 44.87523816, "S": 18.39864105, "O": 36.72612079},
                "atomic percent": {
                    "K": 28.57142857,
                    "S": 14.28571429,
                    "O": 57.14285714,
                },
                "oxide percent": {"K2O": 54.05676836, "SO3": 45.94323164},
            },
            {
                "formula": "(NH 4)2 SO4*H2O",
                "parsed formula": {"N": 2.0, "H": 10.0, "S": 1.0, "O": 5.0},
                "molar mass": 150.149,
                "mass percent": {
                    "N": 18.65746692,
                    "H": 6.71333142,
                    "S": 21.35212356,
                    "O": 53.2770781,
                },
                "atomic percent": {
                    "N": 11.11111111,
                    "H": 55.55555556,
                    "S": 5.55555556,
                    "O": 27.77777778,
                },
                "oxide percent": {
                    "NO2": 35.09929733,
                    "H2O": 34.36114777,
                    "SO3": 30.5395549,
                },
            },
            {
                "formula": "(K0.6Na0.4)2[S]O4",
                "parsed formula": {"K": 1.2, "Na": 0.8, "S": 1.0, "O": 4.0},
                "molar mass": 161.3656,
                "mass percent": {
                    "K": 29.07534196,
                    "Na": 11.39772046,
                    "S": 19.86792724,
                    "O": 39.65901035,
                },
                "atomic percent": {
                    "K": 17.14285714,
                    "Na": 11.42857143,
                    "S": 14.28571429,
                    "O": 57.14285714,
                },
                "oxide percent": {
                    "K2O": 35.02419351,
                    "Na2O": 15.36362149,
                    "SO3": 49.612185,
                },
            },
        ]

    # set of tests for formula string validity checks
    def test_empty_formula(self) -> None:
        string = ""
        self.assertRaises(ValueError, lambda: ChemicalFormula(string))

    def test_invalid_character(self) -> None:
        string = "猫H2O"
        self.assertRaises(
            InvalidCharacter, lambda: ChemicalFormula(string).parsed_formula
        )

    def test_brackets_balanced(self) -> None:
        string = "(NH42SO4"
        self.assertRaises(
            BracketsNotPaired, lambda: ChemicalFormula(string).parsed_formula
        )

    def test_is_adduct_one(self) -> None:
        string = "(NH4)2SO4*H2O*K2SO4"
        self.assertRaises(
            MoreThanOneAdduct, lambda: ChemicalFormula(string).parsed_formula
        )
        string2 = "(NH4)2SO4*H2O·K2SO4"
        self.assertRaises(
            MoreThanOneAdduct, lambda: ChemicalFormula(string2).parsed_formula
        )

    def test_are_atoms_legal(self) -> None:
        string = "Ca3(JO4)2"
        self.assertRaises(NoSuchAtom, lambda: ChemicalFormula(string).parsed_formula)

    def test_formula_wrong_type(self) -> None:
        string = 3
        self.assertRaises(TypeError, lambda: ChemicalFormula(string))

    def test_rounding_wrong_type(self) -> None:
        string = "H2O"
        self.assertRaises(TypeError, lambda: ChemicalFormula(string, "3"))

    def test_rounding_order_less_than_zero(self) -> None:
        string = "H2O"
        self.assertRaises(
            ValueError, lambda: ChemicalFormula(string, rounding_order=-1)
        )

    # set of tests to check ChemicalFormula properties
    def test_parser(self) -> None:
        for item in self.test_data:
            self.formula = ChemicalFormula(item.get("formula"))
            self.assertEqual(self.formula.parsed_formula, item.get("parsed formula"))

    def test_molar_mass(self) -> None:
        for item in self.test_data:
            self.formula = ChemicalFormula(item.get("formula"))
            self.assertEqual(self.formula.molar_mass, item.get("molar mass"))

    def test_mass_percent(self) -> None:
        for item in self.test_data:
            self.formula = ChemicalFormula(item.get("formula"))
            self.assertEqual(self.formula.mass_percent, item.get("mass percent"))

    def test_atomic_percent(self) -> None:
        for item in self.test_data:
            self.formula = ChemicalFormula(item.get("formula"))
            self.assertEqual(self.formula.atomic_percent, item.get("atomic percent"))

    def test_oxide_percent(self) -> None:
        for item in self.test_data:
            self.formula = ChemicalFormula(item.get("formula"))
            self.assertEqual(self.formula.oxide_percent, item.get("oxide percent"))

    def cleanup_files(self) -> None:
        listdir = os.listdir()
        for item in listdir:
            if item == "CSC_formula_test.txt" or item == "CSC_formula_test.json":
                os.remove(item)
        return

    # exports tests
    def test_txt_export(self) -> None:
        string = self.test_data[0].get("formula")
        self.formula = ChemicalFormula(string)
        self.formula.export_to_txt(filename="CSC_formula_test.txt")

    def test_json_serialization(self) -> None:
        string = self.test_data[0].get("formula")
        self.formula = ChemicalFormula(string)
        dictionary = json.loads(self.formula.as_json())
        output = {
            "formula": "H2O",
            "parsed formula": {"H": 2.0, "O": 1.0},
            "molar mass": 18.015,
            "mass percent": {"H": 11.1907, "O": 88.8093},
            "atomic percent": {"H": 66.6667, "O": 33.3333},
            "oxide percent": {"H2O": 100.0},
        }
        self.assertEqual(dictionary, output)

    def test_json_export(self) -> None:
        string = self.test_data[0].get("formula")
        self.formula = ChemicalFormula(string)
        self.formula.export_to_json(filename="CSC_formula_test.json")

    def tearDown(self) -> None:
        self.cleanup_files()
        return super().tearDown()


if __name__ == "__main__":
    unittest.main()



import numpy as np
import scipy.linalg
import gc

# import cupy as cp # in case of CuPy combinatorial function
from .chemutils import find_gcd, find_lcm
from fractions import Fraction

class Balancer:
    """A class for  balancing chemical equations automatically by different matrix methods
    
    Currently implemented: Thorne algorithm (see :meth:`inv_algorithm` method for details),
    Risteski general pseudo-inverse algorithm (see :meth:`gpinv_algorithm` method for details),
    Risteski partial pseudo-inverse algorithm (see :meth:`ppinv_algorithm` method for details),
    and naive combinational search algorithm (see :meth:`comb_algorithm` for details).
    Class takes two matrices: matrix of reactants and matrix of products of chemical reaction, derived
    from general reaction matrix. Matrices are in the form of NumPy 2D array.

    Note:
        Why using `scipy.linalg.pinv <https://docs.scipy.org/doc/scipy/reference/generated/scipy.linalg.pinv.html>`_,
        when `numpy.linalg.pinv <https://numpy.org/doc/stable/reference/generated/numpy.linalg.pinv.html>`_ is doing
        the same thing and does not require the whole scipy import?

        There are some peculiar reaction cases where (especially for :meth:`ppinv_algorithm` method) the results for
        numpy.linalg.pinv differs from system to syste (np version, OS, python version etc.). My understanding is that
        the cause of this behaviour lies in small differences for pinv algorithm in numpy C-libraries and BLAS-libraries,
        hence the difference. To avoid this, more consistent scipy.linalg.pinv method used.

    Argumetns:
        reactant_matrix (np.array): matrix of reactants property generated by :class:`chemsynthcalc.chemical_reaction.ChemicalReaction` class
        
        product_matrix (np.array): matrix of products property generated by :class:`chemsynthcalc.chemical_reaction.ChemicalReaction` class
        
        rounding_order (int): coefficients rounding precision
        
        intify (bool): determines whether the coefficients should be integers
        
        try_comb (bool): flag, which determines whether an attempt will be made to equalize the reaction using the combinatorial method
    
    Attributes:
        reaction_matrix (np.array): sum of reactant_matrix and product_matrix
        coef_limit (int): max integer coefficient for :meth:`intify_coefficients`
    """

    def __init__(
        self,
        reactant_matrix: np.array,
        product_matrix: np.array,
        rounding_order: int,
        intify: bool,
        try_comb: bool,
        max_comb: int = 1e8,
    ) -> None:
        self.reactant_matrix: np.array = reactant_matrix
        self.product_matrix: np.array = product_matrix
        self.reaction_matrix: np.array = np.hstack(
            (self.reactant_matrix, self.product_matrix)
        )
        self.rounding_order: int = rounding_order
        self.try_comb: bool = try_comb
        self.max_comb: int = max_comb
        self.intify: bool = intify
        self.coef_limit: int = 1000000

    def calculate_coefficients_inv(self) -> list:
        """High-level function call to calculate coefficients using inverse Thorne algorithm.

        Returns:
            list: list of calculated coefficients
            None: if cannot calculate
        
        Example:
            >>> reaction = ChemicalReaction("KMnO4+HCl=MnCl2+Cl2+H2O+KCl")
            >>> Balancer(reaction.reactant_matrix, reaction.product_matrix, 8, True, True).calculate_coefficients_inv()
            [2, 16, 2, 5, 8, 2]
        """
        try:
            coefficients = self.inv_algorithm()
            if self.intify:

                intified_coefficients = self.intify_coefficients(
                    coefficients, self.coef_limit
                )
                if self.is_reaction_balanced(
                    self.reactant_matrix, self.product_matrix, intified_coefficients
                ):
                    return intified_coefficients
                elif self.is_reaction_balanced(
                    self.reactant_matrix, self.product_matrix, coefficients
                ):
                    return [int(i) if i.is_integer() else i for i in coefficients]

            else:
                if self.is_reaction_balanced(
                    self.reactant_matrix, self.product_matrix, coefficients
                ):
                    return [int(i) if i.is_integer() else i for i in coefficients]
                else:
                    return None
        except Exception:
            return None

    def calculate_coefficients_gpinv(self) -> list:
        """High-level function call to calculate coefficients using general Risteski algorithm.
    
        Returns:
            list: list of calculated coefficients
            None: if cannot calculate
        
        Example:
            >>> reaction = ChemicalReaction("KMnO4+HCl=MnCl2+Cl2+H2O+KCl")
            >>> Balancer(reaction.reactant_matrix, reaction.product_matrix, 8, True, True).calculate_coefficients_gpinv()
            [2, 16, 2, 5, 8, 2]
        """
        try:
            coefficients = self.gpinv_algorithm()

            if self.intify:
                intified_coefficients = self.intify_coefficients(
                    coefficients, self.coef_limit
                )
                if self.is_reaction_balanced(
                    self.reactant_matrix, self.product_matrix, intified_coefficients
                ):
                    return intified_coefficients
                elif self.is_reaction_balanced(
                    self.reactant_matrix, self.product_matrix, coefficients
                ):
                    return coefficients

            else:
                if self.is_reaction_balanced(
                    self.reactant_matrix, self.product_matrix, coefficients
                ):
                    return coefficients
                else:
                    return None
        except Exception:
            return None

    def calculate_coefficients_ppinv(self) -> list:
        """High-level function call to calculate coefficientsusing partial Risteski algorithm.

        Returns:
            list: list of calculated coefficients
            None: if cannot calculate

        Example:
            >>> reaction = ChemicalReaction("KMnO4+HCl=MnCl2+Cl2+H2O+KCl")
            >>> Balancer(reaction.reactant_matrix, reaction.product_matrix, 8, True, True).calculate_coefficients_ppinv()
            [2, 16, 2, 5, 8, 2]
        """
        try:
            coefficients = self.ppinv_algorithm()

            if self.intify:
                intified_coefficients = self.intify_coefficients(
                    coefficients, self.coef_limit
                )

                if self.is_reaction_balanced(
                    self.reactant_matrix, self.product_matrix, intified_coefficients
                ):
                    return intified_coefficients
                elif self.is_reaction_balanced(
                    self.reactant_matrix, self.product_matrix, coefficients
                ):
                    return coefficients

            else:
                if self.is_reaction_balanced(
                    self.reactant_matrix, self.product_matrix, coefficients
                ):
                    return coefficients
                else:
                    return None
        except Exception:
            return None

    def calculate_coefficients_comb(self) -> list:
        """High-level function call to calculate coefficients using combinatorial algorithm.

        Returns:
            list: list of calculated coefficients
            None: if cannot calculate

        Example:
            >>> reaction = ChemicalReaction("KMnO4+HCl=MnCl2+Cl2+H2O+KCl")
            >>> Balancer(reaction.reactant_matrix, reaction.product_matrix, 8, True, True).calculate_coefficients_comb()
            [2, 16, 2, 5, 8, 2]
        """
        try:
            coefficients = self.comb_algorithm(max_number_of_iterations=self.max_comb)
            return coefficients
        except Exception:
            return None

    def calculate_coefficients_auto(self) -> list:
        """Automatically balance the chemical reaction.
        
        A method that tries to automatically balance the chemical reaction.
        by sequentially applying the three calculation methods to the reaction matrix:
        a Thorne algorithm :meth:`inv_algorithm`, two Risteski algorithms
        :meth:`ppinv_algorithm` and :meth:`gpinv_algorithm` and, if try_comb flag == True,
        a combinatorial algorithm :meth:`comb_algorithm`.

        Returns:
            tuple: list of calculated coefficients and str of algorithm used
            None: if cannot calculate
        
        Examples:
            >>> reaction = ChemicalReaction("KMnO4+HCl=MnCl2+Cl2+H2O+KCl")
            >>> Balancer(reaction.reactant_matrix, reaction.product_matrix, 8, True, True).calculate_coefficients_auto()
            ([2, 16, 2, 5, 8, 2], 'inverse')
            >>> reaction = ChemicalReaction("H2O2+KNO3+H2SO4=K2SO4+NO+H2O+O2")
            >>> Balancer(reaction.reactant_matrix, reaction.product_matrix, 8, True, True).calculate_coefficients_auto()
            ([73, 118, 59, 59, 118, 132, 125], 'general pseudoinverse')
            >>> reaction = ChemicalReaction("NH4ClO4+HNO3+HCl=HClO4+NOCl+N2O+N2O3+H2O+Cl2")
            >>> Balancer(reaction.reactant_matrix, reaction.product_matrix, 8, True, True).calculate_coefficients_auto()
            ([64, 167, 137, 80, 43, 64, 30, 240, 39], 'general pseudoinverse')
        """
        # try inv
        coefficients = self.calculate_coefficients_inv()
        if coefficients and not any(x <= 0 for x in coefficients):
            return coefficients, "inverse"
        # if failed, try gpinv
        coefficients = self.calculate_coefficients_gpinv()
        if coefficients and not any(x <= 0 for x in coefficients):
            return coefficients, "general pseudoinverse"
        # if failed, try ppinv
        coefficients = self.calculate_coefficients_ppinv()
        if coefficients and not any(x <= 0 for x in coefficients):
            return coefficients, "partial pseudoinverse"
        # if all failed, and try_comb, try comb
        if self.try_comb:
            coefficients = self.calculate_coefficients_comb()
            if coefficients and not any(x <= 0 for x in coefficients):
                return coefficients, "combinatorial"

        print("Cannot automatically balance this reaction")
        return

    @staticmethod
    def is_reaction_balanced(
        reactant_matrix: np.array,
        product_matrix: np.array,
        coefficients: list,
        tolerance: float = 1e-8,
    ) -> bool:
        """Is reaction balanced with specified coefficients.

        Checks if reaction is balanced by multiplying reactant matrix and product matrix
        by respective coefficient vector. Method is static to call it outside of balancer
        instance.

        Arguments:
            reactant_matrix (np.array): matrix of reactants property generated by :class:`chemsynthcalc.chemical_reaction.ChemicalReaction` class
        
            product_matrix (np.array): matrix of products property generated by :class:`chemsynthcalc.chemical_reaction.ChemicalReaction` class
        
            coefficients (list): coefficients

            tolerance(float): tolerance limit for np.allclose function
    
        Returns:
            bool: True if balanced within tolerance

        Examples:
            >>> reaction = ChemicalReaction("NH4ClO4+HNO3+HCl=HClO4+NOCl+N2O+N2O3+H2O+Cl2")
            >>> Balancer.is_reaction_balanced(reaction.reactant_matrix, reaction.product_matrix, [64, 167, 137, 80, 43, 64, 30, 240, 39])
            True
            >>> reaction = ChemicalReaction("H2+O2=H2O")
            >>> Balancer.is_reaction_balanced(reaction.reactant_matrix, reaction.product_matrix, [2,2,2])
            False
        """
        try:
            reactants = np.multiply(
                reactant_matrix.T,
                np.array(coefficients)[: reactant_matrix.shape[1], None],
            )
            products = np.multiply(
                product_matrix.T,
                np.array(coefficients)[reactant_matrix.shape[1] :, None],
            )
            if np.allclose(reactants.sum(axis=0), products.sum(axis=0), rtol=tolerance):
                return True
            else:
                return False
        except Exception:
            return False

    def round_up_coefficients(self, coefficients: list, order: int) -> list:
        """Round up the float coefficient to int.

        Round up the float coefficient if it contains a series
        of the same digit (i.e. 3.00000000009). This function is
        needed to fix inevitable floating point math errors.

        Note:
            There is probably a better and faster way to do it.
        
        Arguments:
            coefficients (list): List of coefficients to round
            order (int): rounding precision
        
        Returns:
            list: list of rounded coefficients
        
        Example:
            >>> reaction = ChemicalReaction("P2O3+HClO3+H2O=H3PO4+HCl")
            >>> Balancer(reaction.reactant_matrix, reaction.product_matrix, 8, True, True).round_up_coefficients([1.500000000000001, 1.0, 4.500000000000002, 3.0000000000000013, 1.0000000000000002], 8))
            [1.5, 1.0, 4.5, 3.0, 1.0]
        """
        for idx, coefficient in enumerate(coefficients):
            coef_decimal = str(coefficient).split(".")[1]
            coefficients[idx] = coefficient
            for i in range(10):
                if str(i) * (order - 2) in coef_decimal:
                    coefficients[idx] = round(coefficient, order)
                    break
        return coefficients

    def intify_coefficients(self, coefficients: list, limit: int) -> list:
        """Reduce the coefficients to integers by finding greatset common divider.
        
        Arguments:
            coefficients (list): List of coefficients to intify
            limit (int): upper limit (max int coef)
        
        Returns:
            list: list of intified coefficients

        Example:
            >>> reaction = ChemicalReaction("P2O3+HClO3+H2O=H3PO4+HCl")
            >>> Balancer(reaction.reactant_matrix, reaction.product_matrix, 8, True, True).intify_coefficients([1.5, 1.0, 4.5, 3.0, 1.0], 100000)
            [3, 2, 9, 6, 2]
        """
        initial_coefficients = coefficients
        frac = [Fraction(x).limit_denominator() for x in coefficients]
        vals = [
            int(
                fr.numerator
                * find_lcm([fr.denominator for fr in frac])
                / fr.denominator
            )
            for fr in frac
        ]
        coefficients = [int(val / find_gcd(vals)) for val in vals]
        if any(x > limit for x in coefficients):
            return initial_coefficients
        return coefficients

    def inv_algorithm(self) -> list:
        """Matrix inverse algorithm for reaction balancing.

        A reaction matrix inverse algorithm proposed by `Thorne <https://arxiv.org/abs/1110.4321>`_.
        The calculation is based on nullity, or dimensionality, of the matrix.

        The algorithm can described in steps:
        1) First, check the reaction matrix shape.
        If number of rows in greater then number of columns, add zero
        columns until the matrix becomes square (Note: this is modification
        of original Thorne method described in article).

        2) If reaction matrix is square (which means that number
        of atoms involved is equal to number of compound) than
        we turn matrix in its row-echelon form by singular value
        decomposition.

        3) Calculation of the nullity of the matrix, which is
        basically number of compounds minus rank of the matrix.

        4) Create matrix augumented by nullity number of rows
        of flipped identity matrix. If any rows are zeroes -
        replace it with identity matrix rows.

        5) Inverse the augumented matrix

        6) Exctract and transpose rightmost column

        7) Normalize this value with absolute min value of vector

        8) Round up float operations errors

        Absolute values of this vector are coefficients for the
        reaction.

        Note:
            While this method works great for reactions with 0 and 1
            nullity, it generally cannot work with nullities 2 and higher.
            Thorne claims that for higher nullities, a nullity number
            of vectors should be extracted, and each of them containes
            a set of correct coefficients. However, if number of rows in
            augumentation flipped identity matrix is 2 or more, one can
            easily see that each vector will contain nullity-1 zeroes,
            therefore they cannot be a correct vector of coefficients.

        Returns:
            list: list of calculated coefficients
        
        Examples:
            >>> reaction = ChemicalReaction("Bi2O3+TiO2=Bi4Ti3O12")
            >>> Balancer(reaction.reactant_matrix, reaction.product_matrix, 8, True, True).inv_algorithm()
            [2.0, 3.0, 1.0]
            >>> reaction = ChemicalReaction("NH3+Br2=N2+NH4Br")
            >>> Balancer(reaction.reactant_matrix, reaction.product_matrix, 8, True, True).inv_algorithm()
            [8.0, 3.0, 1.0, 6.0]
            >>> reaction = ChemicalReaction("Zn+H2S2O7=ZnSO4+H2S+H2O")
            >>> Balancer(reaction.reactant_matrix, reaction.product_matrix, 8, True, True).inv_algorithm()
            [4.0, 2.5, 4.0, 1.0, 1.5]
        """
        reaction_matrix = self.reaction_matrix

        if reaction_matrix.shape[0] > reaction_matrix.shape[1]:
            zero_columns = np.zeros(
                (
                    reaction_matrix.shape[0],
                    reaction_matrix.shape[0] - reaction_matrix.shape[1],
                )
            )
            reaction_matrix = np.hstack((reaction_matrix, zero_columns))

        if reaction_matrix.shape[0] == reaction_matrix.shape[1]:
            p, l, reaction_matrix = np.linalg.svd(reaction_matrix)

        number_of_cols = reaction_matrix.shape[1]
        rank = np.linalg.matrix_rank(reaction_matrix, tol=1e-100)
        nullity = number_of_cols - rank
        augument = np.flip(np.identity(reaction_matrix.shape[1])[:nullity], axis=1)
        augumented_matrix = np.vstack((reaction_matrix, augument))
        if np.where(~augumented_matrix.any(axis=1))[0].size > 0:
            augumented_matrix = augumented_matrix[
                ~np.all(augumented_matrix == 0, axis=1)
            ]
        inversed_matrix = np.linalg.inv(augumented_matrix)
        vector = inversed_matrix[:, -1].T
        vector = np.absolute(np.squeeze(np.asarray(vector)))
        vector = vector[vector != 0]
        coefficients = np.divide(vector, vector.min())
        return self.round_up_coefficients(coefficients.tolist(), self.rounding_order)

    def gpinv_algorithm(self) -> list:
        """Matrix gerenal pseudoinverse algorithm for reaction balancing.

        A reaction matrix pseudoinverse algorithm
        proposed by `Risteski <http://koreascience.or.kr/article/JAKO201314358624990.page>`_.
        There are others articles and methods of chemical
        equation balancing by this author, however, this particular
        algorithm seems to be most convenient for matrix calculations.
        The algorithm can described in steps:

        1) Stack reactant matrix and negative product matrix

        2) Calculate MP pseudoinverse of this matrix

        3) Calculate coefficients by formula:
        x = (I – A+A)a, where x is the coefficients vector,
        I - identity matrix, A+ - MP inverse, A - matrix,
        a - arbitrary vector (in this case, vector of ones).

        Note:
            This method is more general than Thorne method, although it has some
            its own peculiarities. First of all, output of this method is float,
            so, to generate an int coefs list, it needs to be converted, which is
            not always leads to a good result. Secondly, MP pseudoinverse
            is sensetive to row order in the reaction matrix. The rows should
            be ordered by atoms apperances in the reaction string.

        Returns:
            list: list of calculated coefficients
        
        Examples:
            >>> reaction = ChemicalReaction("Bi2O3+TiO2=Bi4Ti3O12")
            >>> Balancer(reaction.reactant_matrix, reaction.product_matrix, 8, True, True).gpinv_algorithm()
            [0.857142857142856, 1.285714285714286, 0.42857142857142894]
            >>> reaction = ChemicalReaction("NH3+Br2=N2+NH4Br")
            >>> Balancer(reaction.reactant_matrix, reaction.product_matrix, 8, True, True).gpinv_algorithm()
            [1.3090909090909086, 0.49090909090909124, 0.16363636363636305, 0.9818181818181823]
            >>> reaction = ChemicalReaction("Zn+H2S2O7=ZnSO4+H2S+H2O")
            >>> Balancer(reaction.reactant_matrix, reaction.product_matrix, 8, True, True).gpinv_algorithm()
            [1.2530120481927716, 0.7831325301204817, 1.253012048192771, 0.313253012048193, 0.4698795180722889]
        """
        matrix = np.hstack((self.reactant_matrix, -self.product_matrix))
        inverse = scipy.linalg.pinv(matrix)
        a = np.ones((matrix.shape[1], 1))
        i = np.identity(matrix.shape[1])
        coefs = (i - inverse @ matrix) @ a
        coefs = coefs.flat[:]
        return coefs.tolist()

    def ppinv_algorithm(self) -> list:
        """Matrix partial pseudoinverse algorithm for reaction balancing.
        
        A reaction matrix pseudoinverse algorithm also
        proposed by `Risteski <https://www.koreascience.or.kr/article/JAKO200802727293429.page>`_.
        The method is founded on virtue of the solution of a
        Diophantine matrix equation by using of a Moore-Penrose
        pseudoinverse matrix.

        The algorithm can described in steps:
        1) Take the Moore-Penrose pseudoinverse of reactant matrix

        2) Create a G matrix in the form of (I-AA^-)B, where
        I is the identity matrix, A is the reactant matrix, A^- is
        the MP pseudoinverse of A and B is the product matrix.

        3) Then, vector y (coefficients of products) is equal to
        (I-G^-G)u.

        4) Vector x (coefficients of reactants) is equal to
        A^-By + (I-A^-A)v, where u and v are columns of ones.

        Note:
            While this algorithm and :meth:`gpinv_algorithm` are
            very similar, there are some differences in output results.
            This method exists mostly for legacy purposes, like balancing
            following reactions according to `Risteski <https://www.koreascience.or.kr/article/JAKO200802727293429.page>`_.

        Returns:
            list: list of calculated coefficients
        
        Examples:
            >>> reaction = ChemicalReaction("PtNH3BrNO3+CuNH3KNO3+BeCO3=C44H50O14.98+Cu(NO3)2+PtO3+Br1.99NO2+K1.97O+BeO+HNO3")
            >>> Balancer(reaction.reactant_matrix, reaction.product_matrix, 8, True, True).ppinv_algorithm()
            [0.13568057806893347, 0.5130378112503738, 1.533818182010022, 0.03485950413659115, 0.5130378112503766, 0.13568057806893571, 0.06818119500951737, 0.26042528489867056, 1.5338181820100212, 0.20317996112835507]
            >>> reaction = ChemicalReaction("PtNH3BrNO3+(Fe(CN)3)4(Fe(CN)2)3+C44H50O15+Cu(NO3)2+K2BeO2=CuNH3KNO3+BeCO3+K3.97Fe(CN)6+PtO2+Br1.96NO2+HNO3")
            >>> Balancer(reaction.reactant_matrix, reaction.product_matrix, 8, True, True).ppinv_algorithm()
            [1.0653080872224086, 0.02084998005190298, 0.03200884162226492, 1.236558074626148, 0.9079895101341948, 1.2365580746261489, 0.9079895101341948, 0.145949860363285, 1.0653080872224083, 0.5435245342971473, 1.0866921189021224]   
            >>> reaction = ChemicalReaction("KAu(CN)2+AgRuAuTe8+Fe2(SO4)3+N2Se4+WO3+Na2CO3+H2CO3+HCl=[Ru(C10H8N2)3]Cl2*6H2O+C4H3AuNa1.96OS7+[WCl4(NSeCl)]2+K3.98Fe(CN)6+Au2O3+TeO3+AgO+NO2")
            >>> Balancer(reaction.reactant_matrix, reaction.product_matrix, 8, True, True).ppinv_algorithm()   
            [3.077274686056925, 0.071072198453605, 0.3865922972433353, 0.03396659573785036, 0.13586638295139766, 0.1623687648421996, 1.117085035989227, 0.8214763116641697, 0.07107219845360468, 0.16568241310428458, 0.06793319147569614, 0.7731845944866663, 1.491332235703123, 0.5685775876288415, 0.07107219845360557, 1.021075422996532]         
        """
        MP_inverse = scipy.linalg.pinv(self.reactant_matrix)
        g_matrix = (
            np.identity(self.reaction_matrix.shape[0])
            - self.reactant_matrix @ MP_inverse
        )
        g_matrix = g_matrix @ self.product_matrix
        y_multiply = scipy.linalg.pinv(g_matrix) @ g_matrix
        y_vector = (np.identity(y_multiply.shape[1]) - y_multiply).dot(
            np.ones(y_multiply.shape[1])
        )
        x_multiply = MP_inverse @ self.reactant_matrix
        x_multiply = (
            np.identity(x_multiply.shape[1]) - x_multiply
        ) + MP_inverse @ self.product_matrix @ y_vector.T
        x_vector = x_multiply[0].T
        coefs = np.squeeze(np.asarray(np.hstack((x_vector, y_vector)))).tolist()
        return coefs

    def comb_algorithm(self, max_number_of_iterations: int = 1e8) -> list:
        """Matrix combinatorial algorithm for reaction balancing.

        Finds a solution solution of a Diophantine matrix equation
        by simply enumerating of all possible solutions of number_of_iterations
        coefficients. The solution space is created by Cartesian product
        (in this case, np.meshgrid function), therefore it is very
        limited by memory. There must a better, clever and fast solution
        to this.

        Important:
            Only for integer coefficients less than 128. Only for reactions
            with total compound count <=10.
            A CuPy GPU-accelerated version of this method is available in the
            source code.

        Note:
            All possible variations of coefficients vectors are
            combinations = max_coefficients**number_of_compounds,
            therefore this method is most effective for reaction with
            small numbers of compounds.
                
        Returns:
            list: list of calculated coefficients
        
        Examples:
        >>> reaction = ChemicalReaction("KMnO4+H2S+H2SO4=S+MnSO4+K2SO4+H2O")
        >>> Balancer(reaction.reactant_matrix, reaction.product_matrix, 8, True, True).comb_algorithm()
        [2, 2, 2, 1, 2, 1, 4]
        >>> reaction = ChemicalReaction("H2O2+KNO3+H2SO4=K2SO4+NO+H2O+O2")
        >>> Balancer(reaction.reactant_matrix, reaction.product_matrix, 8, True, True).comb_algorithm()
        [1, 2, 1, 1, 2, 2, 2]
        >>> reaction = ChemicalReaction("Fe2O3+C=Fe3O4+FeO+Fe+Fe3C+CO+CO2")
        >>> Balancer(reaction.reactant_matrix, reaction.product_matrix, 8, True, True).comb_algorithm()
        [4, 5, 1, 1, 1, 1, 1, 3]
        """
        byte = 127
        number_of_compounds = self.reaction_matrix.shape[1]
        if number_of_compounds > 10:
            raise ValueError("Sorry, this method is only for n of compound <=10")

        number_of_iterations = int(
            max_number_of_iterations ** (1 / number_of_compounds)
        )

        if number_of_iterations > byte:
            number_of_iterations = byte

        trans_reaction_matrix = (self.reaction_matrix).T
        lenght = self.reactant_matrix.shape[1]
        old_reactants = trans_reaction_matrix[:lenght].astype("ushort")
        old_products = trans_reaction_matrix[lenght:].astype("ushort")
        for i in range(2, number_of_iterations + 2):
            cart_array = (np.arange(1, i, dtype="ubyte"),) * number_of_compounds
            permuted = np.array(np.meshgrid(*cart_array), dtype="ubyte").T.reshape(
                -1, number_of_compounds
            )
            filter = np.asarray([i - 1], dtype="ubyte")
            permuted = permuted[np.any(permuted == filter, axis=1)]
            # print("calculating max coef %s of %s" % (i-1, number_of_iterations), end='\r', flush=False)
            reactants_vectors = permuted[:, :lenght]
            products_vectors = permuted[:, lenght:]
            del permuted
            reactants = (old_reactants[None, :, :] * reactants_vectors[:, :, None]).sum(
                axis=1
            )
            products = (old_products[None, :, :] * products_vectors[:, :, None]).sum(
                axis=1
            )
            diff = np.subtract(reactants, products)
            del reactants
            del products
            where = np.where(~diff.any(axis=1))[0]
            if np.any(where):
                if where.shape[0] == 1:
                    idx = where
                else:
                    idx = where[0]
                # print("")
                return np.array(
                    np.concatenate(
                        (
                            reactants_vectors[idx].flatten(),
                            products_vectors[idx].flatten(),
                        )
                    )
                ).tolist()
            gc.collect()
        # print("")
        return None

    '''
    def comb_algorithm(self, max_number_of_iterations:int = 1e8) -> list:
        """Matrix combinatorial algorithm for reaction balancing.

        A CuPy GPU-accelerated version of the same combinatorial
        algorithm. Requires `CUDA toolkit <https://developer.nvidia.com/cuda-toolkit>`_.
        to work. GPU gives around 10x acceleration.

        Finds a solution solution of a Diophantine matrix equation
        by simply enumerating of all possible solutions of number_of_iterations
        coefficients. The solution space is created by Cartesian product
        (in this case, np.meshgrid function), therefore it is very
        limited by memory. There must a better, clever and fast solution
        to this.

        Important:
            Only for integer coefficients less than 128. Only for reactions
            with total compound count <=10.

        Note:
            All possible variations of coefficients vectors are
            combinations = max_coefficients**number_of_compounds,
            therefore this method is most effective for reaction with
            small numbers of compounds.
                
        Returns:
            list: list of calculated coefficients
        
        Examples:
        >>> reaction = ChemicalReaction("KMnO4+H2S+H2SO4=S+MnSO4+K2SO4+H2O")
        >>> Balancer(reaction.reactant_matrix, reaction.product_matrix, 8, True, True).comb_algorithm()
        [2, 2, 2, 1, 2, 1, 4]
        >>> reaction = ChemicalReaction("H2O2+KNO3+H2SO4=K2SO4+NO+H2O+O2")
        >>> Balancer(reaction.reactant_matrix, reaction.product_matrix, 8, True, True).comb_algorithm()
        [1, 2, 1, 1, 2, 2, 2]
        >>> reaction = ChemicalReaction("Fe2O3+C=Fe3O4+FeO+Fe+Fe3C+CO+CO2")
        >>> Balancer(reaction.reactant_matrix, reaction.product_matrix, 8, True, True).comb_algorithm()
        [4, 5, 1, 1, 1, 1, 1, 3]
        """

        byte = 127
        number_of_compounds = self.reaction_matrix.shape[1]
        if number_of_compounds>10:
            raise ValueError("Sorry, this method is for n of compound <=10")

        number_of_iterations = int(max_number_of_iterations**(1/number_of_compounds))

        if number_of_iterations > byte:
            number_of_iterations = byte
    
        trans_reaction_matrix = (self.reaction_matrix).T
        lenght = self.reactant_matrix.shape[1]
        old_reactants = cp.asarray(trans_reaction_matrix[:lenght].astype('ushort'))
        old_products = cp.asarray(trans_reaction_matrix[lenght:].astype('ushort'))
        for i in range(2, number_of_iterations+2):
            cart_array = (cp.arange(1, i, dtype='ubyte'), )*number_of_compounds
            permuted = cp.array(cp.meshgrid(*cart_array), dtype='ubyte').T.reshape(-1,number_of_compounds)
            filter = cp.asarray([i-1], dtype='ubyte')
            permuted = permuted[np.any(permuted==filter, axis=1)]
            print("calculating max coef %s of %s" % (i-1, number_of_iterations), end='\r', flush=True)
            reactants_vectors = cp.asarray(permuted[:, :lenght])
            products_vectors = cp.asarray(permuted[:, lenght:])
            del permuted
            reactants = (old_reactants[None, :, :] * reactants_vectors[:,:, None]).sum(axis=1)
            products = (old_products[None, :, :] * products_vectors[:,:, None]).sum(axis=1)
            diff = cp.subtract(reactants, products)
            del reactants
            del products
            where = cp.where(~diff.any(axis=1))[0]
            if cp.any(where):
                if where.shape[0] == 1:
                    idx = where
                else:
                    idx = where[0]
                print("")
                return cp.array(cp.concatenate((reactants_vectors[idx].flatten(), products_vectors[idx].flatten()))).tolist()
            gc.collect()
        print("")
        print("No solution found")
        return None
    '''


from .periodic_table import periodic_table
from .formula_parser import ChemicalFormulaParser


class MolarMassCalculation:
    """Class for calculation of molar masses and percentages of compounds.
    
    Compounds should be parsed by FormulaParser first.

    Arguments:
        parsed_formula (dict): formula parsed by :class:`chemsynthcalc.formula_parser.ChemicalFormulaParser`

    Attributes:
        atomicWeight (list): list of (atom, weight) tuple formed from :mod:`chemsynthcalc.periodic_table`
        list_of_atoms (list): list of atoms formed from :mod:`chemsynthcalc.periodic_table`
    """

    def __init__(self, parsed_formula: dict) -> None:
        self.atomicWeight = [(x[0], x[1]) for x in periodic_table]
        self.parsed_formula = parsed_formula
        self.list_of_atoms = [x[0] for x in periodic_table]

    def calculate_molar_mass(self) -> float:
        """Calculation of molar mass of compound from atomic masses of atoms in parsed formula.
        
        Atomic masses are taken from :mod:`chemsynthcalc.periodic_table` file.

        Returns:
            float: molar mass (in g/mol)

        Examples:
            >>> MolarMassCalculation({'H':2, 'O':1}).calculate_molar_mass()
            18.015
            >>> MolarMassCalculation({'C':2, 'H':6, 'O':1}).calculate_molar_mass()
            46.069
        """
        atomic_masses_list = []
        for atom, n in self.parsed_formula.items():
            atom_index = self.list_of_atoms.index(atom)
            atomic_masses_list.append(self.atomicWeight[atom_index][1] * n)
        return sum(atomic_masses_list)

    def calculate_mass_percent(self) -> dict:
        """Calculation of mass percents of atoms in parsed formula.

        Returns:
            dict: mass percentages of atoms in formula
        
        Examples:
            >>> MolarMassCalculation({'H':2, 'O':1}).calculate_mass_percent()
            {'H': 11.19067443796836, 'O': 88.80932556203163}
            >>> MolarMassCalculation({'C':2, 'H':6, 'O':1}).calculate_mass_percent()
            {'C': 52.14352384466777, 'H': 13.12813388612733, 'O': 34.72834226920489}
        """
        atomic_masses_list = []
        for atom, n in self.parsed_formula.items():
            atom_index = self.list_of_atoms.index(atom)
            atomic_masses_list.append(self.atomicWeight[atom_index][1] * n)
        molar_mass = self.calculate_molar_mass()
        percents = [atomic / molar_mass * 100 for atomic in atomic_masses_list]
        return dict(zip(self.parsed_formula.keys(), percents))

    def calculate_atomic_percent(self) -> dict:
        """Calculation of atomic percents of atoms in parsed formula.

        Returns:
            dict: atomic percentages of atoms in formula
        
        Examples:
            >>> MolarMassCalculation({'H':2, 'O':1}).calculate_atomic_percent()
            {'H': 66.66666666666666, 'O': 33.33333333333333
            >>> MolarMassCalculation({'C':2, 'H':6, 'O':1}).calculate_atomic_percent()
            {'C': 22.22222222222222, 'H': 66.66666666666666, 'O': 11.11111111111111}
        """
        values = list(self.parsed_formula.values())
        total = sum(values)
        atomic = [i / total * 100 for i in values]
        return dict(zip(self.parsed_formula.keys(), atomic))

    def calculate_oxide_percent(self) -> dict:
        """Calculation of oxide percents in parsed formula.

        Calculation of oxide percents in parsed formula from types
        of oxides declared in periodic table file. This type of data
        is mostly used in XRF spectrometry and mineralogy. The oxide
        percents are calculated by finding the convertion factor between elenent
        and its respective oxide (https://www.geol.umd.edu/~piccoli/probe/molweight.html)
        and normalizing the total sum to 100%. One can change the oxide type
        for certain element in periodic_table file. Theoretically, this function
        should work for other types of binary compound (sulfides, fluorides etc.)
        or even salts, however, modification of this function is required
        (for instance, in case of binary compound, removing X atom
        from list of future compounds should have X as an argument of this function)

        Returns:
            dict: oxide percentages of oxides in formula
        
        Examples:
            >>> MolarMassCalculation({'C':2, 'H':6, 'O':1}).calculate_oxide_percent()
            {'CO2': 61.9570190690046, 'H2O': 38.04298093099541}
            >>> MolarMassCalculation({'Ba':1, 'Ti':1, 'O':3}).calculate_oxide_percent()
            {'BaO': 65.7516917244869, 'TiO2': 34.24830827551309}
        """
        oxide_types = [i[3] for i in periodic_table]
        old_atoms = list(self.parsed_formula.keys())
        old_weights = list(
            MolarMassCalculation(self.parsed_formula).calculate_mass_percent().values()
        )
        # Remove O from list of future oxides
        atoms = []
        weights = []
        if "O" in old_atoms:
            temprow = []
            for j, item in enumerate(old_weights):
                if j != old_atoms.index("O"):
                    temprow.append(item)
            weights.append(temprow)
            for k, item in enumerate(old_atoms):
                if k != old_atoms.index("O"):
                    atoms.append(item)
        else:
            atoms = old_atoms
            weights = [old_weights]
        molar_weights = [
            elements[1]
            for atom in atoms
            for elements in periodic_table
            if atom == elements[0]
        ]
        indxs = [
            periodic_table.index(elements)
            for atom in atoms
            for elements in periodic_table
            if atom == elements[0]
        ]
        oxides = [oxide_types[indx] for indx in indxs]
        parsed_oxides = [
            ChemicalFormulaParser(formula).parse_formula() for formula in oxides
        ]
        molar_masses = [
            MolarMassCalculation(oxide).calculate_molar_mass()
            for oxide in parsed_oxides
        ]
        convertion_factors = []
        for i, atom in enumerate(atoms):
            convertion_factor = (
                molar_masses[i] / molar_weights[i] / parsed_oxides[i].get(atom)
            )
            convertion_factors.append(convertion_factor)
        oxide_percents = []
        for row in weights:
            oxide_percent = [row[x] * convertion_factors[x] for x in range(len(row))]
            oxide_percents.append([x / sum(oxide_percent) * 100 for x in oxide_percent])

        return dict(zip(oxides, oxide_percents[0]))



import re
import numpy as np
from re import compile
from functools import cached_property, lru_cache
from .chemical_formula import ChemicalFormula
from .reaction_matrix import ChemicalReactionMatrix
from .reaction_balance import Balancer
from .chem_output import ReactionOutput
from .chemutils import stripe_formula_from_coefficients, arguments_type_checking
from .chem_errors import (
    BadCoeffiecients,
    NoSuchMode,
    NoSeparator,
    InvalidCharacter,
    ReactantProductDifference,
    ReactionNotBalanced,
    NoSuchAlgorithm,
)


class ChemicalReaction:
    """A class that represents a chemical reaction and do operations on it.

    There are three calculation modes:

    1. "force" mode is used when user enters coefficients
    in the reaction string and want masses to be calculated
    whether the reaction is balanced or not.

    2. "check" mode is the same as force, but with reaction
    balance checks.

    3. "balance" mode  tries to automatically calculate
    coefficients from the reaction string.
    
    Important:
        Unlike other properties of this class, `coefficients`
        property can be set directly.

    Arguments:
        reaction (str): a reaction string
        target (int): index of target compound (0 by default, or first compound in the products)
        mode (str): coefficients calculation mode
        target_mass (float): desired mass of target compound (in grams)
        rounding_order (int): value of rounding precision (8 by default)
        try_comb (bool): flag, which determines whether an attempt will be made to equalize the reaction using the combinatorial method in "balance" mode

    Attributes:
        allowed_symbols (str): characters that allowed in the reaction string
        possible_reaction_separators (list): list of characters that can be used as reactants-products separator
        reactant_separator (str): character that can be used as compounds separator
        types_of_modes (list): allowed types of calculation modes
        algorithm (str): currently used calculation algorithm
    
    Raises:
        ValueError: if rounding order <=0
        NoSuchMode: if `mode` is not in `types_of_modes`
        ValueError: if `target` is out of bounds of product list
        ValueError: if `target` < 0
        ValueError: if `target_mass` <=0

    Examples:
        >>> ChemicalReaction("H2+O2=H2O")
        H2+O2=H2O
        >>> ChemicalReaction("2H2+O2=2H2O", mode="balance").coefficients
        [2, 1, 2]
        >>> ChemicalReaction("2H2+O2=2H2O", mode="check").masses
        [0.11190674, 0.88809326, 1.0]
    """

    def __init__(
        self,
        reaction: str = "",
        target: int = 0,
        mode: str = "balance",
        target_mass: float = 1.0,
        rounding_order: int = 8,
        try_comb: bool = False,
    ) -> None:
        arguments_type_checking(reaction, str)
        arguments_type_checking(target, int)
        arguments_type_checking(mode, str)
        arguments_type_checking(target_mass, int, float)
        arguments_type_checking(rounding_order, int)
        arguments_type_checking(try_comb, bool)

        if rounding_order > 0:
            self.rounding_order: int = rounding_order
        else:
            raise ValueError("rounding order <= 0")

        self.allowed_symbols: str = r"[^a-zA-Z0-9.({[)}\]*·•=<\->→⇄+ ]"
        # separators order is important
        self.possible_reaction_separators: list[str] = [
            "==",
            "=",
            "<->",
            "->",
            "<>",
            ">",
            "→",
            "⇄",
        ]
        self.reactant_separator: str = "+"
        self.types_of_modes: list[str] = ["force", "check", "balance"]
        self.temp_reaction: str = reaction.replace(" ", "")
        self.algorithm: str = ""
        self.try_comb: bool = try_comb
        if mode in self.types_of_modes:
            self.mode: str = mode
        else:
            raise NoSuchMode(
                "There is no mode %s! Please choose between force, check or balance modes"
                % mode
            )
        if target >= 0:
            if target < len(self.products):
                self.target: int = target + len(self.reactants)
            else:
                raise ValueError("Target should be in range of products number")
        else:
            raise ValueError("Target < 0")

        if target_mass > 0:
            self.target_mass: float = target_mass
        else:
            raise ValueError("Target mass cannot be 0 or lower")

    def __repr__(self):
        return str(self.reaction)

    def __str__(self):
        return str(self.reaction)

    @property
    @lru_cache
    def reaction(self) -> str:
        """Initial reaction

        Returns:
            str: Initial reaction string with validity check.

        Raises:
            ValueError: if the reaction string is empty
            InvalidCharacter: if some of characters in the string are invalid
        
        Example:
            >>> ChemicalReaction("KMnO4+HCl=MnCl2+Cl2+H2O+KCl").reaction
            KMnO4+HCl=MnCl2+Cl2+H2O+KCl
        """
        if self.temp_reaction == "":
            raise ValueError("No reaction!")

        if self._is_reaction_string_valid() == True:
            return self.temp_reaction
        else:
            raise InvalidCharacter(
                "Invalid character(s) in the reaction string: %s"
                % self._is_reaction_string_valid()
            )

    @property
    @lru_cache
    def separator(self) -> str:
        """Reactants-product separator.

        Returns:
            str: Separator between reactants and products of chemical reaction.
        
        Raises:
            NoSeparator: if no separator was found in the reaction string

        Example:
            >>> ChemicalReaction("KMnO4+HCl=MnCl2+Cl2+H2O+KCl").separator
            =
        """
        if self._reaction_contains_separator():
            return self._reaction_contains_separator()
        else:
            raise NoSeparator("No separator in reaction")

    @property
    @lru_cache
    def reactants(self) -> list:
        """List of initially sptlitted reactants (left side of the reaction string).
        
        Returns:
            list: Formulas on the left side are splitted by reactant_separator (+) 
            and includes initial coefficients (in case of force or check modes).

        Example:
            >>> ChemicalReaction("KMnO4+HCl=MnCl2+Cl2+H2O+KCl").reactants
            ['KMnO4', 'HCl']
        """
        return self.reaction.split(self.separator)[0].split(self.reactant_separator)

    @property
    @lru_cache
    def products(self) -> list:
        """List of initially sptlitted products (right side of the reaction string).
        
        Returns:
            list: Formulas on the right side that have been splitted
            by reactant_separator (+) and includes initial coefficients (in case of force or check modes).

        Example:
            >>> ChemicalReaction("KMnO4+HCl=MnCl2+Cl2+H2O+KCl").products
            ['MnCl2', 'Cl2', 'H2O', 'KCl']
        """
        return self.reaction.split(self.separator)[1].split(self.reactant_separator)

    @property
    @lru_cache
    def compounds(self) -> list:
        """List of all initially sptlitted products (left side and right side).

        Returns:
            list: Sum of `reactants` and `products`

        Example:
            >>> ChemicalReaction("KMnO4+HCl=MnCl2+Cl2+H2O+KCl").compounds
            ['KMnO4', 'HCl', 'MnCl2', 'Cl2', 'H2O', 'KCl']
        """
        return self.reactants + self.products

    @property
    @lru_cache
    def initial_coefficients(self) -> list:
        """Initial coefficients

        The coefficients are not equal to 1 in case if 
        they have been entered (generally the case for 
        force and check modes).

        Returns:
            list: List of initial coefficients striped from the compounds.

        Example:
            >>> ChemicalReaction("KMnO4+HCl=MnCl2+Cl2+H2O+KCl").initial_coefficients
            [1.0, 1.0, 1.0, 1.0, 1.0, 1.0]
        """
        return [
            stripe_formula_from_coefficients(compound)[0] for compound in self.compounds
        ]

    @property
    @lru_cache
    def formulas(self) -> list:
        """Decomposition of list of formulas from the reaction string

        Returns:
            list: Every formula is striped from coefficient and become
            the :class:`chemsynthcalc.chemical_formula.ChemicalFormula` object.
        
        Example:
            >>> ChemicalReaction("KMnO4+HCl=MnCl2+Cl2+H2O+KCl").formulas
            [KMnO4, HCl, MnCl2, Cl2, H2O, KCl]
        """
        striped = [
            stripe_formula_from_coefficients(compound)[1] for compound in self.compounds
        ]
        return [ChemicalFormula(formula) for formula in striped]

    @property
    @lru_cache
    def parsed_formulas(self) -> list:
        """List of parsed formulas of :class:`chemsynthcalc.chemical_formula.ChemicalFormula` objects list.
        
        Example:
            >>> ChemicalReaction("KMnO4+HCl=MnCl2+Cl2+H2O+KCl").parsed_formulas
            [{'K': 1.0, 'Mn': 1.0, 'O': 4.0}, {'H': 1.0, 'Cl': 1.0},{'Mn': 1.0, 'Cl': 2.0}, {'Cl': 2.0}, {'H': 2.0, 'O': 1.0}, {'K': 1.0, 'Cl': 1.0}]
        """
        return [compound.parsed_formula for compound in self.formulas]

    @property
    @lru_cache
    def matrix(self) -> np.array:
        """Chemical reaction matrix.

        The first implementation of reaction matrix method is probably
        belongs to `Blakley <https://doi.org/10.1021/ed059p728>`_. In general,
        a matrix of chemical reaction is composed of coefficients of each
        atom in each compound, giving a 2D array. The matrix composes
        naturally from previously parsed formulas.

        Returns:
            np.array: 2D array of each atom amount in each formula

        Example:
            >>> ChemicalReaction("KMnO4+HCl=MnCl2+Cl2+H2O+KCl").matrix
            [[1. 0. 0. 0. 0. 1.]  (K)
             [1. 0. 1. 0. 0. 0.]  (Mn)
             [4. 0. 0. 0. 1. 0.]  (O)
             [0. 1. 0. 0. 2. 0.]  (H)
             [0. 1. 2. 2. 0. 1.]] (Cl)
        """
        return ChemicalReactionMatrix(self.parsed_formulas).create_reaction_matrix()

    @property
    @lru_cache
    def reactant_matrix(self) -> np.array:
        """Left half of the reaction matrix
        
        Returns:
            np.array: 2D array of each atom amount in each reactant

        Example:
            >>> ChemicalReaction("KMnO4+HCl=MnCl2+Cl2+H2O+KCl").reactant_matrix
            [[1. 0.]  (K)
             [1. 0.]  (Mn)
             [4. 0.]  (O)
             [0. 1.]  (H)
             [0. 1.]] (Cl)
        """
        return self.matrix[:, : len(self.reactants)]

    @property
    @lru_cache
    def product_matrix(self) -> np.array:
        """Right half of the reaction matrix
        
        Returns:
            np.array: 2D array of each atom amount in each product

        Example:
            >>> ChemicalReaction("KMnO4+HCl=MnCl2+Cl2+H2O+KCl").product_matrix
            [[0. 0. 0. 1.]  (K)
             [1. 0. 0. 0.]  (Mn)
             [0. 0. 1. 0.]  (O)
             [0. 0. 2. 0.]  (H)
             [2. 2. 0. 1.]] (Cl)
        """
        return self.matrix[:, len(self.reactants) :]

    @property
    @lru_cache
    def molar_masses(self) -> list:
        """List of molar masses (in g/mol)
        
        Returns:
            list: List of molar masses of each compound in `compounds`
        
        Example:
            >>> ChemicalReaction("KMnO4+HCl=MnCl2+Cl2+H2O+KCl").molar_masses
            [158.032, 36.458, 125.838, 70.9, 18.015, 74.548]
        """
        return [compound.molar_mass for compound in self.formulas]

    @cached_property
    def coefficients(self) -> list:
        """Coefficients of the chemical reaction.
        
        There are 3 possible modes that method can run:

        1) force mode is when coefficients are entered by user in reaction
        string and the calculation and the calculation takes place regardless
        of reaction balance (it gives warning if reaction is not balanced);

        2) check mode is basically the force mode but it will raise a
        error if reaction is not balanced and will not calculate masses);

        3) balance mode uses one of three auto-balancing aglorithms described
        in detail in :class:`chemsynthcalc.reaction_balance.Balancer` class.

        In the fist two cases, the coefficients are just stripped from original
        formulas entered by user. In case of balance mode, coefficients are
        calculated.

        Important:
            Unlike other properties of this class, this
            property can be set directly.

        Raises:
            ReactionNotBalanced: if reaction is not balanced in "check" mode.

        Returns:
            list: list of reaction coefficients
            None: if it cannot balance the reaction

        Examples:
            >>> ChemicalReaction("KMnO4+HCl=MnCl2+Cl2+H2O+KCl", mode="balance").coefficients
            [2, 16, 2, 5, 8, 2]
            >>> reaction = ChemicalReaction("KMnO4+HCl=MnCl2+Cl2+H2O+KCl", mode="check")
            >>> reaction.coefficients = [2, 16, 2, 5, 8, 2]
            >>> reaction.coefficients
            [2, 16, 2, 5, 8, 2]
            >>> ChemicalReaction("2H2+2O2=H2O", mode="force").coefficients
            [2, 2, 1]
        """
        self.check_elements_count()

        if self.mode == "force":
            if not Balancer.is_reaction_balanced(
                self.reactant_matrix, self.product_matrix, self.initial_coefficients
            ):
                self.algorithm = "user"
            return self.to_integer(self.initial_coefficients)

        elif self.mode == "check":
            if Balancer.is_reaction_balanced(
                self.reactant_matrix, self.product_matrix, self.initial_coefficients
            ):
                self.algorithm = "user"
                return self.to_integer(self.initial_coefficients)
            else:
                raise ReactionNotBalanced("This reaction is not balanced!")

        elif self.mode == "balance":
            try:
                balance = Balancer(
                    self.reactant_matrix,
                    self.product_matrix,
                    self.rounding_order,
                    True,
                    self.try_comb,
                ).calculate_coefficients_auto()
                self.algorithm = balance[1]
                return balance[0]
            except Exception:
                return None

    @property
    @lru_cache
    def normalized_coefficients(self) -> list:
        """List of coefficients normalized on target compound.
        
        target coefficient = 1.0

        Returns:
            list: Normalized coefficients.
        
        Example:
            >>> ChemicalReaction("KMnO4+HCl=MnCl2+Cl2+H2O+KCl").normalized_coefficients
            [1, 8, 1, 2.5, 4, 1]
        """
        if self.coefficients_check(self.coefficients):
            normalized_coefficients = [
                coef / self.coefficients[self.target] for coef in self.coefficients
            ]
            return [
                int(i) if i.is_integer() else round(i, self.rounding_order)
                for i in normalized_coefficients
            ]

    @property
    def is_balanced(self) -> bool:
        """Is reaction balanced.

        Returns:
            bool: True if the reaction is balanced with current coefficients
        
        Example:
            >>> ChemicalReaction("KMnO4+HCl=MnCl2+Cl2+H2O+KCl").is_balanced
            True
        """
        balance = Balancer.is_reaction_balanced(
            self.reactant_matrix, self.product_matrix, self.coefficients
        )
        return balance

    @property
    @lru_cache
    def final_reaction(self) -> str:
        """Final representasion of reaction with coefficients.
        
        Returns:
            str: string of final reaction
        
        Example:
            >>> ChemicalReaction("KMnO4+HCl=MnCl2+Cl2+H2O+KCl").final_reaction
            "2KMnO4+16HCl=2MnCl2+5Cl2+8H2O+2KCl"
        """
        return self.generate_final_reaction(self.coefficients)

    @property
    @lru_cache
    def final_reaction_normalized(self) -> str:
        """Final representasion of reaction with normalized coefficients.
        
        Returns:
            str: String of final normalized reaction
        
        Example:
            >>> ChemicalReaction("KMnO4+HCl=MnCl2+Cl2+H2O+KCl").final_reaction_normalized
            "KMnO4+8HCl=MnCl2+2.5Cl2+4H2O+KCl"
        """
        return self.generate_final_reaction(self.normalized_coefficients)

    @property
    @lru_cache
    def masses(self) -> list:
        """List of masses of compounds (in grams).
        
        List of masses of the of formulas in reaction
        calculated with coefficients obtained by any of 3 methods.
        Calculates masses by calculating amount of substance nu (nu=mass/molar mass).
        Coefficients of reaction are normalized to the target. After nu of target compound is
        calculated, it broadcasted to other compound (with respect to its coefficients).

        Returns:
            list: List of masses of compound

        Examples:
            >>> ChemicalReaction("KMnO4+HCl=MnCl2+Cl2+H2O+KCl").masses
            [1.25583687, 2.31777365, 1.0, 1.40855703, 0.57264101, 0.59241247]
            >>> ChemicalReaction("KMnO4+HCl=MnCl2+Cl2+H2O+KCl", target_mass=2.0).masses
            [2.51167374, 4.63554729, 2.0, 2.81711407, 1.14528203, 1.18482493]
        """
        if self.coefficients_check(self.coefficients):
            nu = self.target_mass / self.molar_masses[self.target]
            masses = [
                round(molar * nu * self.normalized_coefficients[i], self.rounding_order)
                for i, molar in enumerate(self.molar_masses)
            ]
            return masses

    @property
    @lru_cache
    def output_results(self) -> dict:
        """Collection of every output of calculated chemical reaction properties.
        
        Returns:
            dict: All outputs collected in one dictionary.
        
        Example:
            >>> ChemicalReaction("KMnO4+HCl=MnCl2+Cl2+H2O+KCl").output_results
            {'initial reaction': 'KMnO4+HCl=MnCl2+Cl2+H2O+KCl',
            'reaction matrix': array([[1., 0., 0., 0., 0., 1.],
            [1., 0., 1., 0., 0., 0.],
            [4., 0., 0., 0., 1., 0.],
            [0., 1., 0., 0., 2., 0.],
            [0., 1., 2., 2., 0., 1.]]),
            'mode': 'balance',
            'formulas': [KMnO4, HCl, MnCl2, Cl2, H2O, KCl],
            'coefficients': [2, 16, 2, 5, 8, 2],
            'normalized coefficients': [1, 8, 1, 2.5, 4, 1],
            'algorithm': 'inverse', 
            'is balanced': True, 
            'final reaction': '2KMnO4+16HCl=2MnCl2+5Cl2+8H2O+2KCl', 
            'final reaction normalized': 'KMnO4+8HCl=MnCl2+2.5Cl2+4H2O+KCl', 
            'molar masses': [158.032, 36.458, 125.838, 70.9, 18.015, 74.548], 
            'target': MnCl2, 
            'masses': [1.25583687, 2.31777365, 1.0, 1.40855703, 0.57264101, 0.59241247]}
        """
        output = {
            "initial reaction": self.reaction,
            "reaction matrix": self.matrix,
            "mode": self.mode,
            "formulas": self.formulas,
            "coefficients": self.coefficients,
            "normalized coefficients": self.normalized_coefficients,
            "algorithm": self.algorithm,
            "is balanced": self.is_balanced,
            "final reaction": self.final_reaction,
            "final reaction normalized": self.final_reaction_normalized,
            "molar masses": self.molar_masses,
            "target": self.formulas[self.target],
            "masses": self.masses,
        }
        return output

    def _is_reaction_string_valid(self) -> list:
        """Checks if the reaction string is valid for parsing.
        
        The reaction string is valid if it's not contains any 
        characters that are not allowed.

        Returns:
            bool: True if all characters are allowed
            list: List of characters that are not allowed
        
        Examples:
            >>> ChemicalReaction("KMnO4+HCl=MnCl2+Cl2+H2O+KCl")._is_reaction_string_valid()
            True
            >>> ChemicalReaction("KMnO4+HCl=фMnCl2+Cl2+H2O+KCl")._is_reaction_string_valid()
            chemsynthcalc.chem_errors.InvalidCharacter: Invalid character(s) in the reaction string: ['ф']
        """
        search = compile(self.allowed_symbols).findall
        if search(self.temp_reaction):
            return search(self.temp_reaction)
        else:
            return True

    def _reaction_contains_separator(self) -> str:
        """Checks if the reaction string contains one of reactants-products separators:

        Returns:
            str: Separator
            None: if no separator was found
        
        Examples:
            >>> ChemicalReaction("KMnO4+HCl⇄MnCl2+Cl2+H2O+KCl")._reaction_contains_separator()
            ⇄
            >>> ChemicalReaction("KMnO4+HClMnCl2+Cl2+H2O+KCl")._reaction_contains_separator()
            chemsynthcalc.chem_errors.NoSeparator: No separator in reaction
        """
        for separator in self.possible_reaction_separators:
            if self.temp_reaction.find(separator) != -1:
                if self.temp_reaction.split(separator)[1] != "":
                    return separator
        return

    def coefficients_check(self, coefficients: list) -> bool:
        """Checking the coefficients.

        Arguments:
            coefficients (list): list of coefficients

        Raises:
            BadCoeffiecients: If Coefficients are None
            BadCoeffiecients: If 0 or -x in coefficients
            BadCoeffiecients: If number of coefficients is not equal to number of columns in reaction matrix
        
        Returns:
            bool: True, if coefficients are good.
        
        Examples:
            >>> reaction = ChemicalReaction("KMnO4+HCl=MnCl2+Cl2+H2O+KCl")
            >>> reaction.coefficients = [2, 16, 2, 5, 8, 2]
            >>> reaction.coefficients_check(reaction.coefficients)
            True
            >>> reaction.coefficients = [2, 16, 2, 5, 8]
            >>> reaction.coefficients_check(reaction.coefficients)
            chemsynthcalc.chem_errors.BadCoeffiecients: number of coefficients should be equal to 6
            >>> reaction.coefficients = None
            >>> reaction.coefficients_check(reaction.coefficients)
            chemsynthcalc.chem_errors.BadCoeffiecients: Coefficients are None
            >>> reaction.coefficients = [2, 16, 2, 5, 8, 0]
            >>> reaction.coefficients_check(reaction.coefficients)
            chemsynthcalc.chem_errors.BadCoeffiecients: 0 or -x in coefficients
        """
        if coefficients == None:
            raise BadCoeffiecients("Coefficients are None")
        elif any(x <= 0 for x in coefficients):
            raise BadCoeffiecients("0 or -x in coefficients")
        elif len(coefficients) != self.matrix.shape[1]:
            raise BadCoeffiecients(
                "number of coefficients should be equal to %s" % self.matrix.shape[1]
            )

        return True

    @lru_cache
    def check_elements_count(self) -> set:
        """Checks if all elements are present in both sides of reaction.

        Doesn't work in the `force` mode!

        Returns:
            set: Set of difference of atoms in left and right sides of reaction.

        Examples:
            >>> ChemicalReaction("KMnO4+HCl=MnCl2+Cl2+H2O+KCl").check_elements_count()
            None
            >>> ChemicalReaction("KMnHoO4+HCl=MnCl2+Cl2+H2O+KCl").check_elements_count()
            chemsynthcalc.chem_errors.ReactantProductDifference: Cannot balance this
            reaction, because element(s) {'Ho'} are only in one part of the reaction
        """
        if self.mode != "force":
            reactants = {
                k: v
                for d in self.parsed_formulas[: len(self.reactants)]
                for k, v in d.items()
            }
            products = {
                k: v
                for d in self.parsed_formulas[len(self.reactants) :]
                for k, v in d.items()
            }
            diff = set(reactants.keys()).symmetric_difference(set(products.keys()))
            if diff:
                raise ReactantProductDifference(
                    "Cannot balance this reaction, because element(s) %s are only in one part of the reaction"
                    % diff
                )
            else:
                return
        else:
            return

    def to_integer(self, coefficients: list) -> list:
        """Cast a float to integer in a list if it is integer.

        Arguments:
            coefficients (list): list of coefficients

        Returns:
            list: List of intified coefficients.

        Example:
            >>> ChemicalReaction("KMnO4+HCl=MnCl2+Cl2+H2O+KCl").to_integer([1.0, 8.0, 1.0, 2.5, 4.0, 1.0])
            [1, 8, 1, 2.5, 4, 1]
        """
        return [int(i) if i.is_integer() else i for i in coefficients]

    def balance_reaction(
        self, algorithm: str = "inv", intify: bool = True, max_comb: int = 1e8
    ) -> list:
        """High-level function call for all balancing algorithms.
        
        Options availiable are:
        1) `"inv"` - matrix inverse Thorne algorithm
        2) `"gpinv"` - general pseudoinverse Risteski algorithm
        3) `"ppinv"` - partial pseudoinverse Risteski algorithm
        4) `"comb"` - combinatorial algorithm

        Important:
            "comb" algorithm works only for integer coefficients
            less than 127 (see `Balancer.comb_algorithm` for details).

        Arguments:
            algorithm (str) - algortihm choice. "inv" by default.
            intify (bool) - determines whether the coefficients should be integers. True by default.
            max_comb (int) - maximum amount of coefficients vectors hat algorithm goes through. 1e8 by default.

        Returns:
            list: List of obtained coefficients.

        Examples:
            >>> ChemicalReaction("KMnO4+HCl=MnCl2+Cl2+H2O+KCl").balance_reaction(algorithm="inv", intify=True)
            [2, 16, 2, 5, 8, 2]
            >>> ChemicalReaction("KMnO4+HCl=MnCl2+Cl2+H2O+KCl").balance_reaction(algorithm="gpinv", intify=False)
            [0.19607843137254966, 1.568627450980392, 0.1960784313725499, 0.49019607843137436, 0.7843137254901958, 0.19607843137254813]
        """
        arguments_type_checking(algorithm, str)
        arguments_type_checking(intify, bool)
        arguments_type_checking(max_comb, int, float)

        self.check_elements_count()

        if self.mode != "balance":
            raise ValueError("Reaction balancing is only available in balance mode")

        avaliable_algorithms: list[str] = ["inv", "gpinv", "ppinv", "comb"]
        if algorithm not in avaliable_algorithms:
            raise NoSuchAlgorithm(
                "There is no algorithm %s! Please choose between inv, gpinv, ppinv and comb algorithms"
                % algorithm
            )

        if max_comb <= 0:
            raise ValueError("Max combinations <= 0")

        balance = Balancer(
            self.reactant_matrix,
            self.product_matrix,
            self.rounding_order,
            intify,
            self.try_comb,
            max_comb,
        )

        if algorithm == "inv":
            coefficients = balance.calculate_coefficients_inv()
            if coefficients:
                self.algorithm = "inverse"
                return coefficients
            else:
                print("Can't equalize this reaction by inverse algorithm")
                return None

        elif algorithm == "gpinv":
            coefficients = balance.calculate_coefficients_gpinv()
            if coefficients:
                self.algorithm = "general pseudoinverse"
                return coefficients
            else:
                print("Can't equalize this reaction by general pseudoinverse algorithm")
                return None

        elif algorithm == "ppinv":
            coefficients = balance.calculate_coefficients_ppinv()
            if coefficients:
                self.algorithm = "partial pseudoinverse"
                return coefficients
            else:
                print("Can't equalize this reaction by partial pseudoinverse algorithm")
                return None

        elif algorithm == "comb":
            coefficients = balance.calculate_coefficients_comb()
            if coefficients:
                self.algorithm = "combinatorial"
                return coefficients
            else:
                print("Can't equalize this reaction by combinatorial algorithm")
                return None

    def generate_final_reaction(self, coefs: list) -> str:
        """Final reaction string with connotated formulas and calculated coefficients.

        Arguments:
            coefs (list): list of coefficients

        Returns:
            str: String of the final reaction
        
        Example:
            >>> ChemicalReaction("KMnO4+HCl=MnCl2+Cl2+H2O+KCl").generate_final_reaction([2, 16, 2, 5, 8, 2])
            "2KMnO4+16HCl=2MnCl2+5Cl2+8H2O+2KCl"
        """
        final_reaction = []
        for i, compound in enumerate(self.formulas):
            if coefs[i] != 1 or coefs[i] != 1.0:
                final_reaction.append(str(coefs[i]) + str(compound))
            else:
                final_reaction.append(str(compound))
        final_reaction = (self.reactant_separator).join(final_reaction)
        final_reaction = final_reaction.replace(
            self.reactants[-1] + self.reactant_separator,
            self.reactants[-1] + self.separator,
        )
        return final_reaction
    
    def print_results(self, print_rounding_order: int = 4) -> None:
        """Method to print a final result of calculations in terminal.

        Arguments:
            print_rounding_order (int): print precision (4 digits by default)
        
        Returns:
            None
        """
        printing = ReactionOutput(self.output_results).print_results(
            print_rounding_order
        )
        return

    def export_to_txt(
        self, filename: str = "default", print_rounding_order: int = 4
    ) -> None:
        """Method to print a final result of calculations in txt file.

        Arguments:
            filename (str): filename string (should end with .txt)
            print_rounding_order (int): print precision (4 digits by default)
        
        Returns:
            None
        """
        printing = ReactionOutput(self.output_results).export_to_txt(
            filename, print_rounding_order
        )
        return

    def as_json(self, print_rounding_order: int = 4) -> str:
        """Serialization of output into JSON object.

        Argruments:
            print_rounding_order (int): print precision (4 digits by default).
        
        Returns:
            str: JSON object of output results.
        """
        return ReactionOutput(self.output_results).dump_to_json(print_rounding_order)

    def export_to_json(
        self, filename: str = "default", print_rounding_order: int = 4
    ) -> None:
        """Method to print a final result of calculations in JSON file.

        Arguments:
            filename (str): filename string (should end with .json)
            print_rounding_order (int): print precision (4 digits by default)

        Returns:
            None
        """
        printing = ReactionOutput(self.output_results).export_to_json(
            filename, print_rounding_order
        )
        return


from .formula_parser import ChemicalFormulaParser
from .molar_mass import MolarMassCalculation
from .chem_output import FormulaOutput
from .chemutils import arguments_type_checking
from functools import lru_cache


class ChemicalFormula:
    """A base class for representing a single chemical formula.

    It constructs with a formula string and calculates properties:
    parsed formula, molar mass, mass percent, atomic percent,
    oxide percent  from this string using 
    :class:`chemsynthcalc.formula_parser.ChemicalFormulaParser` and
    :class:`chemsynthcalc.molar_mass.MolarMassCalculation`.

    Arguments:
        formula (srt): string of chemical formula
        rounding_order (int): value of rounding precision (8 by default)

    Raises:
        ValueError: if the input string is empty
        ValueError: if rounding order <= 0

    Examples:
        >>> ChemicalFormula("H2O")
        H2O
        >>> ChemicalFormula("H2O").molar_mass
        18.015
        >>> ChemicalFormula("H2O").mass_percent
        {'H': 11.19067444, 'O': 88.80932556}
    """

    def __init__(self, formula: str = "", rounding_order: int = 8) -> None:
        arguments_type_checking(formula, str)
        arguments_type_checking(rounding_order, int)
        if formula == "":
            raise ValueError("No formula!")
        self.formula: str = formula.replace(" ", "")
        if rounding_order > 0:
            self.rounding_order: int = rounding_order
        else:
            raise ValueError("rounding order <= 0")

    def __repr__(self) -> str:
        return str(self.formula)

    def __str__(self) -> str:
        return str(self.formula)

    @property
    @lru_cache
    def parsed_formula(self) -> dict:
        """Formula parsed into dict.

        Returns:
            dict: Parsed dictionary representation of formula string
            created by :class:`chemsynthcalc.formula_parser.ChemicalFormulaParser`.

        Example:
            >>> ChemicalFormula("K2SO4").parsed_formula
            {'K': 2.0, 'S': 1.0, 'O': 4.0}
        """
        parsed = ChemicalFormulaParser(self.formula).parse_formula()
        return {k: round(v, self.rounding_order + 3) for k, v in parsed.items()}

    @property
    @lru_cache
    def molar_mass(self) -> float:
        """Molar mass of formula.

        Returns:
            float: `molar mass <https://en.wikipedia.org/wiki/Molar_mass>`_
            of the formula (in g/mol), calculated from
            parsed formula using :class:`chemsynthcalc.molar_mass.MolarMassCalculation`.

        Example:
            >>> ChemicalFormula("K2SO4").molar_mass
            174.252
        """
        return round(
            MolarMassCalculation(self.parsed_formula).calculate_molar_mass(),
            self.rounding_order,
        )

    @property
    @lru_cache
    def mass_percent(self) -> dict:
        """ Mass percents of atoms in formula.

        Returns:
            dict: A mass percent or 
            `relative mass fraction <https://en.wikipedia.org/wiki/Mass_fraction_(chemistry)>`_
            of atoms in parsed chemical formula. The values of
            mass content are in % (with 100% sum), not fraction.
        
        Example:
            >>> ChemicalFormula("K2SO4").mass_percent
            {'K': 44.87523816, 'S': 18.39864105, 'O': 36.72612079}
        """
        output = MolarMassCalculation(self.parsed_formula).calculate_mass_percent()
        return {k: round(v, self.rounding_order) for k, v in output.items()}

    @property
    @lru_cache
    def atomic_percent(self) -> dict:
        """Atomic percents of atoms in formula.

        Returns:
            dict: An atomic percent or 
            `relative mole fraction <https://en.wikipedia.org/wiki/Mole_fraction>`_
            dictionary of atoms in parsed chemical formula.
            The values of mole content are in % (with 100% sum),
            not fraction
        
        Example:
            >>> ChemicalFormula("K2SO4").atomic_percent
            {'K': 28.57142857, 'S': 14.28571429, 'O': 57.14285714}
        """
        output = MolarMassCalculation(self.parsed_formula).calculate_atomic_percent()
        return {k: round(v, self.rounding_order) for k, v in output.items()}

    @property
    @lru_cache
    def oxide_percent(self) -> dict:
        """Oxide percents of metals in formula.

        Returns:
            dict: An oxide percent or 
            `oxide fraction <https://d32ogoqmya1dw8.cloudfront.net/files/introgeo/studio/examples/minex02.pdf>`_ 
            dictionary of atoms in parsed chemical formula. Oxide types are listed
            in :mod:`chemsynthcalc.periodic_table` file and can be changed to
            any oxide formula. The values of oxide content
            are in % (with 100% sum), not fraction.

        Example:
            >>> ChemicalFormula("K2SO4").oxide_percent
            {'K2O': 54.05676836, 'SO3': 45.94323164}
        """
        output = MolarMassCalculation(self.parsed_formula).calculate_oxide_percent()
        return {k: round(v, self.rounding_order) for k, v in output.items()}

    @property
    @lru_cache
    def output_results(self) -> dict:
        """Dictionary of calculation resulst output for class.

        Returns:
            dict: Output dictionary for all properties listed above.

        Example:
            >>> ChemicalFormula("H2O").output_results
            {'formula': 'H2O', 
            'parsed formula': {'H': 2.0, 'O': 1.0}, 
            'molar mass': 18.015, 
            'mass percent': {'H': 11.19067444, 'O': 88.80932556}, 
            'atomic percent': {'H': 66.66666667, 'O': 33.33333333}, 
            'oxide percent': {'H2O': 100.0}}
        """
        output: dict = {
            "formula": self.formula,
            "parsed formula": self.parsed_formula,
            "molar mass": self.molar_mass,
            "mass percent": self.mass_percent,
            "atomic percent": self.atomic_percent,
            "oxide percent": self.oxide_percent,
        }
        return output

    def print_results(self, print_rounding_order: int = 4) -> None:
        """ Method to print a final result of calculations in terminal.

        Arguments:
            print_rounding_order (int): print precision (4 digits by default)
                    
        Returns:
            None
        """
        printing = FormulaOutput(self.output_results).print_results(
            print_rounding_order
        )
        return

    def export_to_txt(
        self, filename: str = "default", print_rounding_order: int = 4
    ) -> None:
        """Method to print a final result of calculations in a txt file.

        Arguments:
            filename (str): filename string (should end with .txt)
            print_rounding_order (int): print precision (4 digits by default)
        
        Returns:
            None
        """
        printing = FormulaOutput(self.output_results).export_to_txt(
            filename, print_rounding_order
        )
        return

    def as_json(self, print_rounding_order: int = 4) -> str:
        """Serialization of output into JSON object.

        Arguments:
            print_rounding_order (int): print precision (4 digits by default)
        
        Returns:
            str: A JSON-type object of results output.
        """
        return FormulaOutput(self.output_results).dump_to_json(print_rounding_order)

    def export_to_json(
        self, filename: str = "default", print_rounding_order: int = 4
    ) -> None:
        """Method to print a final result of calculations in JSON file.

        Arguments:
            filename (str): filename string (should end with .json)
            print_rounding_order (int): print precision (4 digits by default)
        
        Returns:
            None
        """
        printing = FormulaOutput(self.output_results).export_to_json(
            filename, print_rounding_order
        )
        return



# Configuration file for the Sphinx documentation builder.
#
# For the full list of built-in configuration values, see the documentation:
# https://www.sphinx-doc.org/en/master/usage/configuration.html

# -- Project information -----------------------------------------------------
# https://www.sphinx-doc.org/en/master/usage/configuration.html#project-information

project = 'chemsynthcalc'
copyright = '2022, Egor Syrov'
author = 'Egor Syrov'
release = '1.0.3'

# -- General configuration ---------------------------------------------------
# https://www.sphinx-doc.org/en/master/usage/configuration.html#general-configuration
import sys
import os
import sphinx_rtd_theme


sys.path.append("./src")
import chemsynthcalc

    # If extensions (or modules to document with autodoc) are in another directory,
    # add these directories to sys.path here. If the directory is relative to the
    # documentation root, use os.path.abspath to make it absolute, like shown here.

extensions = ['sphinx.ext.autodoc', 
'sphinx.ext.autosummary', 
'sphinx.ext.napoleon', 
'sphinx_rtd_theme',]
autodoc_member_order = 'bysource'
# Napoleon settings
napoleon_google_docstring = True
napoleon_numpy_docstring = False
napoleon_include_init_with_doc = False
napoleon_include_private_with_doc = False
napoleon_include_special_with_doc = True
napoleon_use_admonition_for_examples = False
napoleon_use_admonition_for_notes = False
napoleon_use_admonition_for_references = False
napoleon_use_ivar = False
napoleon_use_param = True
napoleon_use_rtype = True
napoleon_preprocess_types = False
napoleon_type_aliases = None
napoleon_attr_annotations = True

master_doc = 'index'

templates_path = ['_templates']
exclude_patterns = []

language = 'English'

# -- Options for HTML output -------------------------------------------------
# https://www.sphinx-doc.org/en/master/usage/configuration.html#options-for-html-output

html_theme = "sphinx_rtd_theme"
html_static_path = ['_static']


from .periodic_table import periodic_table
import numpy as np

class ChemicalReactionMatrix:
    """Class that creates a reaction matrix from parsed formulas of the parsed reaction string.
    
    Parameters:
        parsed_all (list): list of all parsed formulas
    
    Attributes:
        merged_dict (dict): merged total dict of all parsed formulas
        electronegativity (list): list of (atom, electronegativity) tuples formed from :mod:`chemsynthcalc.periodic_table` file
        electronegativity_pattern (list): list of atoms sorted by their Pauling electronegativities
        list_of_atoms (list): list of atoms from :mod:`chemsynthcalc.periodic_table` file.
    """

    def __init__(self, parsed_all: list) -> None:
        self.parsed_all: list = parsed_all
        self.merged_dict: dict = {k: v for d in self.parsed_all for k, v in d.items()}
        self.electronegativity: list = [(x[0], x[2]) for x in periodic_table]
        self.electronegativity_pattern: list = [
            atom[0] for atom in sorted(self.electronegativity, key=lambda tup: tup[1])
        ]
        self.list_of_atoms = [x[0] for x in periodic_table]

    def create_reaction_matrix(self) -> np.array:
        """Create chemical reaction matrix.
        
        Method that creates a chemical matrix of reaction from list of
        parsed compound. Initially contains a species type (atom label)
        at the start of each row of matrix (might be used in case of future
        expansion of matrix representation in print). Retruns 2D
        NumPy array (since Matrix class is no longer recommended
        https://numpy.org/doc/stable/reference/generated/numpy.matrix.html).

        Returns:
            np.array: 2D array of chemical reaction matrix 
            (amounts of all atoms in all compounds of reaction)

        Example:
            >>> parsed = ChemicalReaction("H2+O2=H2O").parsed_formulas
            >>> ChemicalReactionMatrix(parsed).create_reaction_matrix()
            [[2. 0. 2.]
             [0. 2. 1.]]
        """
        elements = list(self.merged_dict.keys())
        elements_vector = np.array(elements, dtype="str_").T
        matrix = []
        for i, element in enumerate(elements):
            row = []
            for parsed in self.parsed_all:
                if element in parsed.keys():
                    row.append(parsed.get(elements[i]))
                else:
                    row.append(0)
            matrix.append(row)
        return np.array(matrix)



"""
ChemSynthCalc
=====
# Chemical synthesis calculator
Python package for calculating the masses of substances required for chemical synthesis directly from the reaction string. It includes solutions for all intermidiate steps, including chemical formula parsing, molar mass calculation and reaction balancing with different matrix methods.

## Example use
Let's say that we need to prepare 3 grams of [YBCO](https://en.wikipedia.org/wiki/Yttrium_barium_copper_oxide) by solid-state synthesis from respective carbonates. The reaction string will look something like this (to simplify, let's leave it without oxygen nonstoichiometry):

```Python
from chemsynthcalc import ChemicalReaction

reaction_string = "BaCO3 + Y2(CO3)3 + CuCO3 + O2 → YBa2Cu3O7 + CO2"
```

Now, we can create a chemical reaction object of `ChemicalReaction` class, which will be used in the calculation. We need to specify arguments for our particular case:

```Python

reaction = ChemicalReaction(
    reaction = reaction_string, # our reaction string
    target = 0, # number of target compound in product list
    target_mass = 3, # desired mass of target compound,
    mode = "balance" # mode of coefficients calculations,
)
```

Now, to perform automatic calculation, all we need to do is to put:

```Python
reaction.print_results(print_rounding_order=4) # assuming we use analytical balances with 4 digits presicion
```

And we get our ouput in the console:

```
initial reaction: BaCO3+Y2(CO3)3+CuCO3+O2→YBa2Cu3O7+CO2
reaction matrix:
[[1. 0. 0. 0. 2. 0.]
 [1. 3. 1. 0. 0. 1.]
 [3. 9. 3. 2. 7. 2.]
 [0. 2. 0. 0. 1. 0.]
 [0. 0. 1. 0. 3. 0.]]
mode: balance
coefficients: [8, 2, 12, 1, 4, 26]
normalized coefficients: [2, 0.5, 3, 0.25, 1, 6.5]
balanced by algorithm: inverse
is balanced: True
final reaction: 8BaCO3+2Y2(CO3)3+12CuCO3+O2→4YBa2Cu3O7+26CO2
final reaction normalized: 2BaCO3+0.5Y2(CO3)3+3CuCO3+0.25O2→YBa2Cu3O7+6.5CO2
target: YBa2Cu3O7
BaCO3: M = 197.3380 g/mol, m = 1.7773 g
Y2(CO3)3: M = 357.8360 g/mol, m = 0.8057 g
CuCO3: M = 123.5540 g/mol, m = 1.6692 g
O2: M = 31.9980 g/mol, m = 0.0360 g
YBa2Cu3O7: M = 666.1970 g/mol, m = 3.0000 g
CO2: M = 44.0090 g/mol, m = 1.2882 g
```

Now, we got all masses ready for our planned synthesis!

## Documentation
See the docmentation for detailed functionality and all features of the package.

## Features
* Formula parsing

```Python
from chemsynthcalc import ChemicalFormula

>>>ChemicalFormula("C2H5OH").parsed_formula
{'C': 2.0, 'H': 6.0, 'O': 1.0}
```

* Molar mass calculation

```Python
>>>ChemicalFormula("C2H5OH").molar_mass
46.069
```

* [Mass](https://en.wikipedia.org/wiki/Mass_fraction_(chemistry)), [atomic](https://en.wikipedia.org/wiki/Mole_fraction), and [oxide](https://d32ogoqmya1dw8.cloudfront.net/files/introgeo/studio/examples/minex02.pdf) percent calculations with `mass_percent`, `atomic_percent` and `oxide_percent` properties of `ChemicalFormula`.
* Auto-balancing chemical equations by 4 different matrix methods in `"balance"` mode:

```Python
from chemsynthcalc import ChemicalReaction

reaction_string = "K4Fe(CN)6 + KMnO4 + H2SO4 = KHSO4 + Fe2(SO4)3 + MnSO4 + HNO3 + CO2 + H2O"

>>>ChemicalReaction(reaction_string, mode="balance").final_reaction
"10K4Fe(CN)6+122KMnO4+299H2SO4=162KHSO4+5Fe2(SO4)3+122MnSO4+60HNO3+60CO2+188H2O"
```

* Calculation of masses for user-defined coefficients in `"force"` (calculates regardless of balance) and `"check"` (checks if reaction is balanced by user-defined coefficients) modes.
  
```Python
>>>ChemicalReaction("BaCO3+TiO2=BaTiO3", mode="force").masses #we can drop CO2 product and still get masses in this mode. 
[0.84623961, 0.34248308, 1.0]
```

```Python
>>>ChemicalReaction("2H2+O2=2H2O", mode="check").coefficients #we can be sure that reaction is balanced with our coefficients in this mode
[2, 1, 2]
```
Setting the coefficients directly into `ChemicalReaction` instance:
```Python
>>>reaction = ChemicalReaction("H2+O2=H2O", mode="check")
>>>reaction.coefficients = [2,1,2]
>>>reaction.coefficients
[2, 1, 2]
>>>reaction.is_balanced
True
```
* Calculation of coefficients with `ChemicalReaction.balance_reaction()` method individually by each of 4 different algorithms (inverse, general pseudoinverse, partial pseudoinverse and combinatorial algorithms).
* Export of results of both `ChemicalFormula` and `ChemicalReaction` into .txt file (with `.export_to_txt()`), into JSON object (with `.as_json()`) or JSON file (with `.export_to_json()`).

"""
__version__ = "1.0.4"
from .chemical_formula import ChemicalFormula
from .chemical_reaction import ChemicalReaction

def run_test() -> None:
    """Run suite of tests both for :class:`chemsynthcalc.chemical_formula.ChemicalFormula` and :class:`chemsynthcalc.chemical_reaction.ChemicalReaction` classes.
    """
    import unittest
    import os
    
    loader = unittest.TestLoader()
    dir_name = os.path.dirname(os.path.realpath(__file__))
    start_dir =  dir_name+"/tests"
    suite = loader.discover(start_dir)
    runner = unittest.TextTestRunner()
    runner.run(suite)


import sys
import time
import json
import numpy as np


class FormulaOutput:
    """Outputs for :class:`chemsynthcalc.chemical_formula.ChemicalFormula`

    A collection of methods for outputing results of
    :class:`chemsynthcalc.chemical_formula.ChemicalFormula`. Default filenames of
    output files are "CSC_formula" with formula
    and nanosecond since the epoch.
    
    Arguments:
        chemical_formula_output (dict): output of :class:`chemsynthcalc.chemical_formula.ChemicalFormula`
    """

    def __init__(self, chemical_formula_output: dict) -> None:
        self.output: dict = chemical_formula_output
        self.original_stdout = sys.stdout

    def __print_stream(self, print_rounding_order: int) -> None:
        """Prints of results that can be streamed into stdout.

        Arguments:
            print_rounding_order (int): print precision
        
        Returns:
            None
        """
        print("formula:", self.output.get("formula"))
        print("parsed formula:", self.output.get("parsed formula"))
        print("molar mass:", round(self.output.get("molar mass"), print_rounding_order))
        print(
            "mass percent:",
            {
                k: round(v, print_rounding_order)
                for k, v in self.output.get("mass percent").items()
            },
        )
        print(
            "atomic percent:",
            {
                k: round(v, print_rounding_order)
                for k, v in self.output.get("atomic percent").items()
            },
        )
        print(
            "oxide percent:",
            {
                k: round(v, print_rounding_order)
                for k, v in self.output.get("oxide percent").items()
            },
        )
        return

    def print_results(self, print_rounding_order: int) -> None:
        """Print results in the terminal.
        
        Arguments:
            print_rounding_order (int): print precision
                
        Returns:
            None
        """
        sys.stdout = self.original_stdout
        self.__print_stream(print_rounding_order)
        return

    def export_to_txt(self, filename: str, print_rounding_order: int) -> None:
        """Print results in the txt file and saves it.
        
        Arguments:
            filename (str): filename string (should end with .txt)
            print_rounding_order (int): print precision
                
        Returns:
            None
        """
        if filename == "default":
            filename = "CSC_formula_%s_%s.txt" % (
                self.output.get("formula:"),
                time.time_ns(),
            )

        with open(filename, "w") as file:
            sys.stdout = file
            self.__print_stream(print_rounding_order)
        sys.stdout = self.original_stdout
        return

    def dump_to_json(self, print_rounding_order: int) -> str:
        """JSON serialization.
        
        Arguments:
            print_rounding_order (int): print precision

        Returns:
            str: A JSON-type object of results output.
        """
        mod_output = self.output.copy()
        mod_output.update(
            {
                "mass percent": {
                    k: round(v, print_rounding_order)
                    for k, v in self.output.get("mass percent").items()
                }
            }
        )
        mod_output.update(
            {
                "atomic percent": {
                    k: round(v, print_rounding_order)
                    for k, v in self.output.get("atomic percent").items()
                }
            }
        )
        mod_output.update(
            {
                "oxide percent": {
                    k: round(v, print_rounding_order)
                    for k, v in self.output.get("oxide percent").items()
                }
            }
        )
        return json.dumps(mod_output, ensure_ascii=False)

    def export_to_json(self, filename: str, print_rounding_order: int) -> None:
        """Dump output dict into JSON flie.
 
        Arguments:
            filename (str): filename string (should end with .json)
            print_rounding_order (int): print precision
                
        Returns:
            None
        """
        if filename == "default":
            filename = "CSC_formula_%s_%s.json" % (
                self.output.get("formula"),
                time.time_ns(),
            )

        with open(filename, "w", encoding="utf-8") as file:
            json_obj = json.loads(self.dump_to_json(print_rounding_order))
            json.dump(json_obj, file, ensure_ascii=False)
        return


class ReactionOutput:
    """Outputs for :class:`chemsynthcalc.chemical_reaction.ChemicalReaction`
    
    A collection of methods for outputing results of
    :class:`chemsynthcalc.chemical_reaction.ChemicalReaction`. Default filenames of
    output files are "CSC_reaction" with target
    compound and nanosecond since the epoch.

    Arguments:
        chemical_formula_output (dict): output of :class:`chemsynthcalc.chemical_reaction.ChemicalReaction`
    """

    def __init__(self, chemical_reaction_output: dict) -> None:
        self.output: dict = chemical_reaction_output
        self.original_stdout = sys.stdout

    def __print_stream(self, print_rounding_order: int) -> None:
        """Prints of results that can be streamed into stdout.

        Arguments:
            print_rounding_order (int): print precision
        
        Returns:
            None
        """
        print("initial reaction:", self.output.get("initial reaction"))
        print("reaction matrix:")
        print(self.output.get("reaction matrix"))
        print("mode:", self.output.get("mode"))
        print("coefficients:", self.output.get("coefficients"))
        print("normalized coefficients:", self.output.get("normalized coefficients"))
        print("balanced by algorithm: %s " % self.output.get("algorithm"))
        print("is balanced:", self.output.get("is balanced"))
        print("final reaction:", self.output.get("final reaction"))
        print(
            "final reaction normalized:", self.output.get("final reaction normalized")
        )
        print("target:", self.output.get("target"))
        formulas = self.output.get("formulas")
        molar_masses = self.output.get("molar masses")
        masses = self.output.get("masses")
        for i, formula in enumerate(formulas):
            print(
                "%s: M = %s g/mol, m = %s g"
                % (
                    formula,
                    "%.{0}f".format(print_rounding_order)
                    % round(molar_masses[i], print_rounding_order),
                    "%.{0}f".format(print_rounding_order)
                    % round(masses[i], print_rounding_order),
                )
            )
        return

    def print_results(self, print_rounding_order: int) -> None:
        """Print results in the terminal.
        
        Arguments:
            print_rounding_order (int): print precision
                
        Returns:
            None
        """
        sys.stdout = self.original_stdout
        self.__print_stream(print_rounding_order)
        return

    def export_to_txt(self, filename: str, print_rounding_order: int) -> None:
        """Print results in the txt file and saves it.
        
        Arguments:
            filename (str): filename string (should end with .txt)
            print_rounding_order (int): print precision
                
        Returns:
            None
        """
        if filename == "default":
            filename = "CSC_reaction_%s_%s.txt" % (
                self.output.get("target"),
                time.time_ns(),
            )
        with open(filename, "w") as file:
            sys.stdout = file
            self.__print_stream(print_rounding_order)
        sys.stdout = self.original_stdout
        return

    def dump_to_json(self, print_rounding_order: int) -> str:
        """JSON serialization.
        
        Arguments:
            print_rounding_order (int): print precision

        Returns:
            str: A JSON-type object of results output.
        """
        mod_output = self.output.copy()
        mod_output.update(
            {"reaction matrix": np.array2string(self.output.get("reaction matrix"))}
        )
        str_formulas = [str(formula) for formula in self.output.get("formulas")]
        mod_output.update({"formulas": str_formulas})
        mod_output.update({"target": str(self.output.get("target"))})
        mod_output.update(
            {
                "molar masses": [
                    round(mass, print_rounding_order)
                    for mass in self.output.get("molar masses")
                ]
            }
        )
        mod_output.update(
            {
                "masses": [
                    round(mass, print_rounding_order)
                    for mass in self.output.get("masses")
                ]
            }
        )
        return json.dumps(mod_output, ensure_ascii=False)

    def export_to_json(self, filename: str, print_rounding_order: int) -> None:
        """Dump output dict into JSON flie.
 
        Arguments:
            filename (str): filename string (should end with .json)
            print_rounding_order (int): print precision
                
        Returns:
            None
        """
        if filename == "default":
            filename = "CSC_reaction_%s_%s.json" % (
                self.output.get("target"),
                time.time_ns(),
            )

        with open(filename, "w", encoding="utf-8") as file:
            json_obj = json.loads(self.dump_to_json(print_rounding_order))
            json.dump(json_obj, file, ensure_ascii=False)
        return


from math import gcd
from functools import reduce

"""
A module with different utility functions, 
which are not belong to main classes
"""


def arguments_type_checking(argument, *args) -> None:
    """Checks for type of function argument.
    
    Argments:
        argument (Any): Argument of the function in question
        types (Any): Types that allowed for this argument
    
    Raises:
        TypeError: if agrument type is not one of specified types
    
    Returns:
        None
    """
    if not isinstance(argument, args):
        raise TypeError(
            "The %s parameter is wrong type %s. Should be %s"
            % (argument, type(argument), args)
        )


def find_lcm(int_list: list) -> int:
    """Finds least common multiple of the list of integers.

    Arguments:
        int_list (list): list of integers
    
    Returns:
        int: Least common multiple
    
    Example:
        >>> chemsynthcalc.chemutils.find_lcm([10,20,30]
        60
    """
    lcm = 1
    for i in int_list:
        lcm = lcm * i // gcd(lcm, i)
    return lcm


def find_gcd(int_list: list) -> int:
    """Finds greatest common divider of the list of integers.

    Arguments:
        int_list (list): list of integers
    
    Returns:
        int: greatest common divider
    
    Example:
        >>> chemsynthcalc.chemutils.find_gcd([10,20,30]
        10
    """
    x = reduce(gcd, int_list)
    return x


def stripe_formula_from_coefficients(formula) -> tuple:
    """A function to stripe the initially parsed formulas from chemical reaction string.
    
    Arguments:
        formula (str): a formula string

    Returns:
        tuple: (float, str) tuple of (coefficient, reaction)
    
    Example:
        >>> chemsynthcalc.chemutils.stripe_formula_from_coefficients("4.3H3PO4")
        (4.3, 'H3PO4')
    """
    if not formula[0].isdigit():
        return 1.0, formula
    else:
        coef = []
        for i, symbol in enumerate(formula):
            if symbol.isdigit() or symbol == ".":
                coef.append(symbol)
            else:
                return float("".join(coef)), formula[i:]

"""
    def intify_coefficients_pinv(self, coefficients:list) -> list:
        '''
        A function to reduce the coefficients to integers by finding 
        greatset common divider.
        '''
        floats = [i for i in coefficients if type(i)==float]
        maximum_lenght = max([len(str(i).split('.')[1]) for i in floats])
        coefficients = [int(i*maximum_lenght*10) for i in coefficients]
        divider = reduce(gcd, coefficients)
        coefficients = [int(i/divider) for i in coefficients]
        return coefficients

"""



from re import findall, match, compile
from collections import Counter
from .periodic_table import periodic_table
from .chem_errors import (
    NoSuchAtom,
    InvalidCharacter,
    MoreThanOneAdduct,
    BracketsNotPaired,
)

class ChemicalFormulaParser:
    """Parser of chemical formulas.
    
    Methods of this class take string of compound chemical formula
    and turn it to dict of atoms as keys and their coefficients as values.

    Arguments:
        formula (str): formula string

    Attributes:
        atom_regex (str): regular expression for finding atoms in formula
        coefficient_regex (str): regular expression for atoms amounts in formula
        atom_and_coefficient_regex (str): atom_regex+coefficient_regex
        opener_brackets (str): opener brackets variations
        closer_brackets (str): closer brackets variations
        adduct_symbols (str): symbols for adduct notation (water of crystallization most often)
        allowed_symbols (str): regular expression for all symbols allowed in the formula string
        list_of_atoms (list): list of all atoms in the periodic table
    """

    def __init__(self, formula: str) -> None:
        self.atom_regex: str = "([A-Z][a-z]*)"
        self.coefficient_regex: str = "((\d+(\.\d+)?)*)"
        self.atom_and_coefficient_regex: str = self.atom_regex + self.coefficient_regex
        self.opener_brackets: str = "({["
        self.closer_brackets: str = ")}]"
        self.adduct_symbols: str = "*·•"
        self.allowed_symbols = r"[^A-Za-z0-9.({[)}\]*·•]"
        self.formula: str = formula
        self.list_of_atoms: list = [x[0] for x in periodic_table]

    def __transform_adduct(self, formula: str) -> str:
        """Transform adduct notation in formula to general brackets notation.

        Arguments:
            formula (str): initial formula string
        
        Returns:
            str: New formula with adduct in bracket notation
        """
        transformed_formula = formula
        for i, token in enumerate(formula):
            if token in self.adduct_symbols:
                m = match(self.coefficient_regex, formula[i + 1 :]).group(0)
                if m:
                    coef = str(m)
                else:
                    coef = ""
                transformed_formula = (
                    formula[:i] + "(" + formula[i + 1 + len(coef) :] + ")" + str(coef)
                )
        return transformed_formula

    def __dictify(self, tuples: tuple) -> dict:
        """Transform tuples of tuples to a dict of atoms.

        Arguments:
            tuples (tuple): tuple of tuples of atoms

        Returns:
            dict: dictionary of atoms
        """
        res = dict()
        for atom, n, m, k in tuples:
            try:
                res[atom] += float(n or 1)
            except KeyError:
                res[atom] = float(n or 1)
        return res

    def __fuse(self, mol1: dict, mol2: dict, w: int = 1) -> dict:
        """Fuse 2 dicts representing molecules. 
        
        Arguments:
            mol1 (dict): dict of atoms 1
            mol2 (dict): dict of atoms 2
            w (int): weight

        Returns:
            dict: a new fuse dict
        """
        return {
            atom: (mol1.get(atom, 0) + mol2.get(atom, 0)) * w
            for atom in set(mol1) | set(mol2)
        }

    def __parse(self, formula: str) -> dict:
        """Parse the formula string
       
        Recurse on opening brackets to parse the subpart and
        return on closing ones because it is the end of said subpart.
        Formula is the argument of this method due to the complications
        of self. Constructions in recursive functions.

        Arguments:
            formula (str): formula string

        Returns:
            dict: Return the molecule dict and length of parsed part
        """
        q = []
        mol = {}
        i = 0

        while i < len(formula):
            # Using a classic loop allow for manipulating the cursor
            token = formula[i]

            if token in self.closer_brackets:
                # Check for an index for this part
                m = match(self.coefficient_regex, formula[i + 1 :]).group(0)
                if m != "":
                    weight = float(m)
                    i += len(m)
                else:
                    weight = 1
                submol = self.__dictify(
                    findall(self.atom_and_coefficient_regex, "".join(q))
                )
                return self.__fuse(mol, submol, weight), i

            elif token in self.opener_brackets:
                submol, l = self.__parse(formula[i + 1 :])
                mol = self.__fuse(mol, submol)
                # skip the already read submol
                i += l + 1
            else:
                q.append(token)

            i += 1

        # Fuse in all that's left at base level
        return (
            self.__fuse(
                mol,
                self.__dictify(findall(self.atom_and_coefficient_regex, "".join(q))),
            ),
            i,
        )

    def _is_formula_valid(self) -> list:
        """Checks if the formula string is valid for parsing
        
        If it does not contains any characters that are not
        allowed.

        Returns:
            bool: True if all characters are allowed
            list: List of characters that are not allowed
        
        Examples:
            >>> ChemicalFormulaParser("K2SO4*H2O")._is_formula_valid()
            True
            >>> ChemicalFormulaParser("K2SO4ンH2O")._is_formula_valid()
            ['ン']
        """
        search = compile(self.allowed_symbols).findall
        if search(self.formula):
            return search(self.formula)
        else:
            return True

    def are_brackets_balanced(self) -> bool:
        """Check if all sort of brackets come in pairs.

        Returns:
            bool: True if all brackets are in pairs
        
        Examples:
            >>> ChemicalFormulaParser("(K2SO4)*H2O").are_brackets_balanced()
            True
            >>> ChemicalFormulaParser("(K2SO4)*H2O").are_brackets_balanced()
            False
        """
        c = Counter(self.formula)
        bracket_counter = c["["] == c["]"] and c["{"] == c["}"] and c["("] == c[")"]

        return bracket_counter

    def is_adduct_one(self) -> bool:
        """ Check if there is only one adduct in formula

        Returns:
            bool: True if there is only one adduct symbol
        
        Examples:
            >>> ChemicalFormulaParser("K2SO4*H2O").is_adduct_one()
            True
            >>> ChemicalFormulaParser("K2SO4*H2O*CO2").is_adduct_one()
            False
        """
        c = Counter(self.formula)
        i = 0
        for adduct in self.adduct_symbols:
            if c[adduct] > 0:
                i += c[adduct]
        if i <= 1:
            return True
        else:
            return False

    def are_atoms_legal(self, parsed:dict) -> None:
        """Checks if all parsed atom are belong to the periodic table.
        
        Arguments:
            parsed (dict): dictionary of parsed atoms
        
        Raises:
            NoSuchAtom: if one of the parsed atoms are not in the periodic table

        Returns:
            None
        
        Examples:
            >>> ChemicalFormulaParser("K2SO4*H2O").are_atoms_legal({'K': 2.0, 'S': 1.0, 'O': 5.0, 'H': 2.0})
            None
            >>> ChemicalFormulaParser("KHuSO4*H2O").are_atoms_legal({'K': 1.0, 'Hu':1.0, 'S': 1.0, 'O': 5.0, 'H': 2.0})
            chemsynthcalc.chem_errors.NoSuchAtom: No atom Hu in the periodic table!
        """
        for atom in list(parsed.keys()):
            if atom not in self.list_of_atoms:
                raise NoSuchAtom("No atom %s in the periodic table!" % atom)
        return

    def parse_formula(self) -> dict:
        """Parse the formula and return a dict with occurences of each atom.

        Raises:
            InvalidCharacter: if some characters in initial string is not in `allowed_characters`
            MoreThanOneAdduct: if there is more than one adduct symbol (i.e. *)
            BracketsNotPaired: if some bracket type is not coming in pairs
        
        Returns:
            dict: Dictionary of parsed atoms

        Examples:
            >>> ChemicalFormulaParser("H2O").parse_formula()
            {'H': 2.0, 'O': 1.0}
            >>> ChemicalFormulaParser("C2H5OH").parse_formula()
            {'C': 2.0, 'H': 6.0, 'O': 1.0}
            >>> ChemicalFormulaParser("(K0.6Na0.4)2SO4*7H2O").parse_formula()
            {'K': 1.2, 'Na': 0.8, 'S': 1.0, 'O': 11.0, 'H': 14.0}
        """
        if self._is_formula_valid() != True:
            raise InvalidCharacter(
                "Invalid character(s) %s in the formula %s"
                % (self._is_formula_valid(), self.formula)
            )

        if not self.is_adduct_one():
            raise MoreThanOneAdduct("More than one adduct in the formula")

        if not self.are_brackets_balanced():
            raise BracketsNotPaired("The brackers are not balanced ![{]$[&?)]}!]")

        transformed = self.__transform_adduct(self.formula)
        parsed = self.__parse(transformed)[0]

        self.are_atoms_legal(parsed)

        # make an ordered atoms dict
        atoms_list = findall(self.atom_regex, self.formula)
        atoms_dict = dict.fromkeys(atoms_list)
        output = {}
        for atom in atoms_dict.keys():
            output[atom] = parsed.get(atom)
        return output